#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2024 Inria

"""Exceptions specific to Pink."""


class PinkError(Exception):
    """Base class for Pink exceptions."""


class ConfigurationError(PinkError):
    """Exception raised when encountering an invalid configuration vector."""


class FrameNotFound(PinkError):
    """Exception raised when a frame is not found in the robot model."""

    def __init__(
        self,
        name: str,
        frames,
    ) -> None:
        """Create exception.

        Args:
            name: Name of the frame.
            frames: List of all robot frames.
        """
        self.name = name
        frame_names = [frame.name for frame in frames]
        self.message = (
            f'Name "{name}" is not a robot frame name in {frame_names}'
        )
        super().__init__(self.message)


class NotWithinConfigurationLimits(PinkError):
    """Exception thrown when a robot configuration violates its limits.

    Attributes:
        joint: Index of the joint in the configuration vector.
        value: Invalid value of the joint.
        lower: Minimum allowed value for this joint.
        upper: Maximum allowed value for this joint.
    """

    joint: int
    value: float
    lower: float
    upper: float

    def __init__(
        self,
        joint: int,
        value: float,
        lower: float,
        upper: float,
    ) -> None:
        """Create exception.

        Args:
            joint: Index of the joint in the configuration vector.
            value: Invalid value of the joint.
            lower: Minimum allowed value for this joint.
            upper: Maximum allowed value for this joint.
        """
        self.joint = joint
        self.value = value
        self.lower = lower
        self.upper = upper
        self.message = (
            f"Joint {joint} violates configuration limits "
            f"{lower} <= {value} <= {upper}"
        )
        super().__init__(self.message)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Limits implemented as inequality constraints in the IK problem."""

from .acceleration_limit import AccelerationLimit
from .limit import Limit
from .configuration_limit import ConfigurationLimit
from .velocity_limit import VelocityLimit

__all__ = [
    "AccelerationLimit",
    "ConfigurationLimit",
    "Limit",
    "VelocityLimit",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""JVRC-1 humanoid standing on two feet and reaching with a hand."""

import meshcat_shapes
import numpy as np
import pinocchio as pin
import qpsolvers
from loop_rate_limiters import RateLimiter

import pink
from pink import solve_ik
from pink.tasks import FrameTask

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try `pip install robot_descriptions`"
    )


class WavingPose:
    """Moving target to the wave the right hand."""

    def __init__(self, init: pin.SE3):
        """Initialize pose.

        Args:
            init: Initial transform from the wrist frame to the world frame.
        """
        self.init = init

    def at(self, t):
        """Get waving pose at a given time.

        Args:
            t: Time in seconds.
        """
        T = self.init.copy()
        R = T.rotation
        R = np.dot(R, pin.utils.rpyToMatrix(0.0, 0.0, np.pi / 2))
        R = np.dot(R, pin.utils.rpyToMatrix(0.0, -np.pi, 0.0))
        T.rotation = R
        T.translation[0] += 0.5
        T.translation[1] += -0.1 + 0.05 * np.sin(8.0 * t)
        T.translation[2] += 0.5
        return T


if __name__ == "__main__":
    robot = load_robot_description(
        "jvrc_description", root_joint=pin.JointModelFreeFlyer()
    )
    viz = pin.visualize.MeshcatVisualizer(
        robot.model, robot.collision_model, robot.visual_model
    )
    robot.setVisualizer(viz, init=False)
    viz.initViewer(open=True)
    viz.loadViewerModel()

    configuration = pink.Configuration(robot.model, robot.data, robot.q0)
    viz.display(configuration.q)

    left_foot_task = FrameTask(
        "l_ankle", position_cost=1.0, orientation_cost=3.0
    )
    pelvis_task = FrameTask(
        "PELVIS_S", position_cost=1.0, orientation_cost=0.0
    )
    right_foot_task = FrameTask(
        "r_ankle", position_cost=1.0, orientation_cost=3.0
    )
    right_wrist_task = FrameTask(
        "r_wrist", position_cost=1.0, orientation_cost=3.0
    )
    tasks = [left_foot_task, pelvis_task, right_foot_task, right_wrist_task]

    pelvis_pose = configuration.get_transform_frame_to_world("PELVIS_S").copy()
    pelvis_pose.translation[0] += 0.05
    meshcat_shapes.frame(viz.viewer["pelvis_pose"])
    viz.viewer["pelvis_pose"].set_transform(pelvis_pose.np)
    pelvis_task.set_target(pelvis_pose)

    transform_l_ankle_target_to_init = pin.SE3(
        np.eye(3), np.array([0.1, 0.0, 0.0])
    )
    transform_r_ankle_target_to_init = pin.SE3(
        np.eye(3), np.array([-0.1, 0.0, 0.0])
    )

    left_foot_task.set_target(
        configuration.get_transform_frame_to_world("l_ankle")
        * transform_l_ankle_target_to_init
    )
    right_foot_task.set_target(
        configuration.get_transform_frame_to_world("r_ankle")
        * transform_r_ankle_target_to_init
    )
    pelvis_task.set_target(
        configuration.get_transform_frame_to_world("PELVIS_S")
    )

    right_wrist_pose = WavingPose(
        configuration.get_transform_frame_to_world("r_wrist")
    )

    wrist_frame = viz.viewer["right_wrist_pose"]
    meshcat_shapes.frame(wrist_frame)

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "quadprog" in qpsolvers.available_solvers:
        solver = "quadprog"

    rate = RateLimiter(frequency=200.0)
    dt = rate.period
    t = 0.0  # [s]
    while True:
        # Update task targets
        right_wrist_task.set_target(right_wrist_pose.at(t))
        wrist_frame.set_transform(right_wrist_pose.at(t).np)

        # Compute velocity and integrate it into next configuration
        velocity = solve_ik(configuration, tasks, dt, solver=solver)
        configuration.integrate_inplace(velocity, dt)

        # Visualize result at fixed FPS
        viz.display(configuration.q)
        rate.sleep()
        t += dt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2024 Inria

"""Test task Jacobian matrices against finite differences."""

import unittest

import numpy as np
import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.tasks import FrameTask, JointCouplingTask, PostureTask, Task


class TestJacobians(unittest.TestCase):
    """Test task Jacobian matrices against finite differences."""

    def setUp(self, nb_configs=10, nb_dirs=2):
        """Prepare test fixture."""
        np.random.seed(42)
        random_dq = 2.0 * np.random.random((nb_dirs, 6)) - 1.0
        l2norms = np.sqrt((random_dq * random_dq).sum(axis=1))
        random_dirs = random_dq / l2norms.reshape((nb_dirs, 1))
        random_q = (
            0.1 * np.pi * (2.0 * np.random.random((nb_configs, 6)) - 1.0)
        )
        robot = load_robot_description("ur3_description")
        self.assertEqual(robot.nq, 6)
        self.link = "ee_link"
        self.data = robot.data
        self.model = robot.model
        self.random_dirs = random_dirs
        self.random_q = random_q
        self.robot = robot

    def check_jacobian_finite_diff(self, task: Task, tol: float):
        """Check that a task Jacobian is de/dq by finite differences.

        Args:
            task: Task to test the Jacobian of.
            tol: Test tolerance.
        """

        def e(q):
            configuration = Configuration(self.model, self.data, q)
            return task.compute_error(configuration)

        def J(q):
            configuration = Configuration(self.model, self.data, q)
            return task.compute_jacobian(configuration)

        nq = self.model.nq
        nv = self.model.nv
        for q_0 in self.random_q:
            J_0 = J(q_0)
            e_0 = e(q_0)

            J_finite = np.empty((e_0.shape[0], nv))
            for i in range(nq):
                h = 0.000001
                e_i = np.eye(nq)[i]
                J_finite[:, i] = (e(q_0 + h * e_i) - e_0) / h

            self.assertLess(np.linalg.norm(J_0 - J_finite, ord=np.inf), tol)

    def test_frame_task(self):
        frame_task = FrameTask(
            self.link,
            position_cost=1.0,
            orientation_cost=1.0,
        )
        frame_task.set_target(pin.SE3.Random())
        self.check_jacobian_finite_diff(frame_task, tol=1e-5)

    def test_joint_coupling_task(self):
        configuration = Configuration(self.model, self.data, self.robot.q0)
        joint_coupling_task = JointCouplingTask(
            ["shoulder_lift_joint", "shoulder_pan_joint"],
            [1.0, -1.0],
            100.0,
            configuration,
        )
        self.check_jacobian_finite_diff(joint_coupling_task, tol=1e-6)

    def test_posture_task(self):
        posture_task = PostureTask(cost=1.0)
        posture_task.set_target(self.robot.q0)
        self.check_jacobian_finite_diff(posture_task, tol=1e-6)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Build and solve the inverse kinematics problem."""

from typing import Iterable, Optional, Tuple

import numpy as np
import qpsolvers

from .barriers import Barrier
from .configuration import Configuration
from .limits import Limit
from .tasks import Task


def __compute_qp_objective(
    configuration: Configuration,
    tasks: Iterable[Task],
    damping: float,
    barriers: Optional[Iterable[Barrier]] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    r"""Compute the QP objective function.

    The Hessian matrix :math:`H` and linear vector :math:`c` define the QP
    objective function as:

    .. math::

        \frac{1}{2} \Delta q^T H \Delta q + c^T q

    The configuration displacement :math:`\Delta q` is the output of inverse
    kinematics (we divide it by :math:`\Delta t` to get a commanded velocity).

    Args:
        configuration: Robot configuration to read kinematics from.
        tasks: List of kinematic tasks to fulfill at (weighted) best.
        damping: weight of Tikhonov (everywhere) regularization. Its unit is
            `[cost]^2 / [tangent]` where `[tangent]` is "the" unit of robot
            velocities. Improves numerical stability, but larger values slow
            down all tasks.
        dt: Integration timestep in [s].
        barriers: Collection of barriers.

    Returns:
        Pair :math:`(H, c)` of Hessian matrix and linear vector of the QP
        objective.
    """
    barriers = barriers if barriers is not None else []
    H = damping * configuration.tangent.eye
    c = configuration.tangent.zeros.copy()
    for task in tasks:
        H_task, c_task = task.compute_qp_objective(configuration)
        H += H_task
        c += c_task

    for barrier in barriers:
        H_cbf, c_cbf = barrier.compute_qp_objective(configuration)
        H += H_cbf
        c += c_cbf

    return (H, c)


def __compute_qp_inequalities(
    configuration,
    limits: Optional[Iterable[Limit]],
    dt: float,
    barriers: Optional[Iterable[Barrier]] = None,
) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
    r"""Compute inequality constraints for the quadratic program.

    Args:
        configuration: Robot configuration to read kinematics from.
        limits: Collection of limits.
        dt: Integration timestep in [s].
        barriers: Collection of barriers.

    Returns:
        Pair :math:`(G, h)` of inequality matrix and vector representing the
        inequality :math:`G \Delta q \leq h`, or ``(None, None)`` if there is
        no inequality.

    Notes:
        We trim comparisons to infinity (equivalently: big floats) because some
        solvers don't support it. See for instance
        https://github.com/stephane-caron/pink/issues/10.
    """
    if limits is None:
        configuration_limit = configuration.model.configuration_limit
        velocity_limit = configuration.model.velocity_limit
        limits = (configuration_limit, velocity_limit)
    barriers = barriers if barriers is not None else []
    G_list = []
    h_list = []
    for limit in limits:
        matvec = limit.compute_qp_inequalities(configuration.q, dt)
        if matvec is not None:
            G_list.append(matvec[0])
            h_list.append(matvec[1])
    for barrier in barriers:
        G_barrier, h_barrier = barrier.compute_qp_inequalities(
            configuration, dt
        )
        G_list.append(G_barrier)
        h_list.append(h_barrier)
    if not G_list:
        return None, None
    return np.vstack(G_list), np.hstack(h_list)


def build_ik(
    configuration: Configuration,
    tasks: Iterable[Task],
    dt: float,
    damping: float = 1e-12,
    limits: Optional[Iterable[Limit]] = None,
    barriers: Optional[Iterable[Barrier]] = None,
) -> qpsolvers.Problem:
    r"""Build quadratic program from current configuration and tasks.

    This quadratic program is, in standard form:

    .. math::

        \begin{split}\begin{array}{ll}
            \underset{\Delta q}{\mbox{minimize}} &
                \frac{1}{2} {\Delta q}^T H {\Delta q} + c^T {\Delta q} \\
            \mbox{subject to}
                & G {\Delta q} \leq h
        \end{array}\end{split}

    where :math:`\Delta q` is a vector of joint displacements corresponding to
    the joint velocities :math:`v = {\Delta q} / {\mathrm{d}t}`.

    Args:
        configuration: Robot configuration to read kinematics from.
        tasks: List of kinematic tasks.
        dt: Integration timestep in [s].
        damping: weight of Tikhonov (everywhere) regularization. Its unit is
            :math:`[\mathrm{cost}]^2 / [\mathrm{tangent}]` where
            :math:`[\mathrm{tangent}]` is "the" unit of robot velocities.
            Improves numerical stability, but larger values slow down all
            tasks.
        limits: Collection of limits to enforce. By default, consists of
            configuration and velocity limits (set to the empty list ``[]`` to
            disable limits).
        barriers: Collection of barriers.

    Returns:
        Quadratic program of the inverse kinematics problem.
    """
    P, q = __compute_qp_objective(configuration, tasks, damping, barriers)
    G, h = __compute_qp_inequalities(configuration, limits, dt, barriers)
    problem = qpsolvers.Problem(P, q, G, h)
    return problem


def solve_ik(
    configuration: Configuration,
    tasks: Iterable[Task],
    dt: float,
    solver: str,
    damping: float = 1e-12,
    limits: Optional[Iterable[Limit]] = None,
    barriers: Optional[Iterable[Barrier]] = None,
    safety_break: bool = True,
    **kwargs,
) -> np.ndarray:
    r"""Compute a velocity tangent to the current robot configuration.

    The computed velocity satisfies at (weighted) best the set of kinematic
    tasks given in argument.

    Args:
        configuration: Robot configuration to read kinematics from.
        tasks: List of kinematic tasks.
        dt: Integration timestep in [s].
        solver: Backend quadratic programming (QP) solver.
        damping: weight of Tikhonov (everywhere) regularization. Its unit is
            :math:`[\mathrm{cost}]^2 / [\mathrm{tangent}]` where
            :math:`[\mathrm{tangent}]` is "the" unit of robot velocities.
            Improves numerical stability, but larger values slow down all
            tasks.
        limits: Collection of limits to enforce. By default, consists of
            configuration and velocity limits (set to the empty list ``[]`` to
            disable limits).
        barriers: Collection of barriers functions.
        safety_break: If True, stop execution and raise an exception if
            the current configuration is outside limits. If False, print a
            warning and continue execution.
        kwargs: Keyword arguments to forward to the backend QP solver.

    Returns:
        Velocity :math:`v` in tangent space.

    Raises:
        NotWithinConfigurationLimits: if the current configuration is not
            within limits.

    Note:
        Our Tikhonov damping is isotropic despite tangent velocities not being
        homogeneous. If it helps we can add a tangent-space scaling to damp the
        floating base differently from joint angular velocities.
    """
    configuration.check_limits(safety_break=safety_break)
    problem = build_ik(
        configuration,
        tasks,
        dt,
        damping,
        limits,
        barriers,
    )
    result = qpsolvers.solve_problem(problem, solver=solver, **kwargs)
    Delta_q = result.x
    assert Delta_q is not None
    v: np.ndarray = Delta_q / dt
    return v



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Ivan Domrachev, Simeon Nedelchev

from ..exceptions import PinkError


class NoPositionLimitProvided(PinkError):
    """If neither minimum nor maximum position limits are provided."""


class NegativeMinimumDistance(PinkError):
    """If the minimum distance in body spherical barrier is negative."""


class InvalidCollisionPairs(PinkError):
    """IF the number of collision pairs is invalid."""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Visualization helpers."""

import pinocchio as pin
from pinocchio import visualize


def start_meshcat_visualizer(
    robot: pin.RobotWrapper,
) -> visualize.MeshcatVisualizer:
    """Open a MeshCat visualizer in a web browser.

    Args:
        robot: Robot wrapper with its initial data.

    Returns:
        MeshCat visualizer.
    """
    viz = visualize.MeshcatVisualizer(
        robot.model, robot.collision_model, robot.visual_model
    )
    robot.setVisualizer(viz, init=False)
    viz.initViewer(open=True)
    viz.loadViewerModel()
    return viz



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Ivan Domrachev, Simeon Nedelchev

"""Barrier based on the distance between two frames."""

from typing import Tuple, Union

import numpy as np

from ..configuration import Configuration
from .barrier import Barrier
from .exceptions import NegativeMinimumDistance


class BodySphericalBarrier(Barrier):
    r"""A barrier based on the distance between two frames.

    Defines a barrier function based on the Euclidean distance between two
    specified frames. It allows for the specification of a minimum distance
    threshold.

    The barrier function is defined as:

    .. math::

        h(q) = \|p_1(q) -
        p_2(q)\|^2 - d_{min}^2

    where :math:`p_1(q)` and
    :math:`p_2(q)` are the positions of the two
    frames in the world coordinate system, and :math:`d_{min}`
    is the minimum distance threshold.

    Attributes:
        frames: Tuple of two frame names.
        d_min: Minimum distance threshold.
    """

    frames: Tuple[str, str]
    d_min: float

    def __init__(
        self,
        frames: Tuple[str, str],
        d_min: float,
        gain: Union[float, np.ndarray] = 1.0,
        safe_displacement_gain: float = 3.0,
    ):
        """Initialize the BodySphericalBarrier.

        Args:
            frames: Tuple of two frame names.
            d_min: Minimum distance threshold.
            gain: Barrier gain. Defaults to 1.0.
            safe_displacement_gain: gain for the safe backup displacement
                cost term. Defaults to 3.0.
        """
        if d_min < 0.0:
            raise NegativeMinimumDistance(
                "The minimum distance threshold must be non-negative."
            )
        super().__init__(
            dim=1,
            gain=gain,
            gain_function=lambda h: h / (1 + np.abs(h)),
            safe_displacement_gain=safe_displacement_gain,
        )
        self.frames = frames
        self.d_min = d_min

    def compute_barrier(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the value of the barrier function.

        The barrier function is computed based on the Euclidean distance
        between the two specified frames. It considers the minimum distance
        threshold.

        The barrier function is given by:

        .. math::

            h(q) = \|p_1(q) -
            p_2(q)\|^2 - d_{min}^2

        where :math:`p_1(q)` and
        :math:`p_2(q)` are the positions of the two
        frames in the world coordinate system, and :math:`d_{min}`
        is the minimum distance threshold.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Value of the barrier function
                :math:`h(q)`.
        """
        pos1_world, pos2_world = self._get_frame_positions(configuration)
        return np.array(
            [
                (pos1_world - pos2_world).dot(pos1_world - pos2_world)
                - self.d_min**2
            ]
        )

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the Jacobian matrix of the barrier function.

        The Jacobian matrix is computed based on the position Jacobians of the
        two specified frames. The Jacobians are transformed to align with the
        world coordinate system.

        The Jacobian matrix is given by:

        .. math::

            \frac{\partial h}{\partial q}(q) =
            2(p_1(q) -
            p_2(q))^T \begin{bmatrix}
                \frac{\partial p_1}{\partial q}
                (q) \\
                -\frac{\partial p_2}{\partial q}
                (q)
            \end{bmatrix}

        where :math:`\frac{\partial p_1}{\partial q}(q)`
        and :math:`\frac{\partial p_2}{\partial q}(q)`
        are the position Jacobians of the two frames.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Jacobian matrix
                :math:`\frac{\partial h}{\partial q}(q)`.
        """
        pos1_world, pos2_world = self._get_frame_positions(configuration)
        pos1_jac, pos2_jac = self._get_frame_jacobians(configuration)

        dh_dx = 2 * (pos1_world - pos2_world)
        dx_dq = pos1_jac - pos2_jac
        return dh_dx.T @ dx_dq

    def _get_frame_positions(
        self, configuration: Configuration
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Get the positions of the two frames in the world coordinate system.

        Args:
            configuration: Robot configuration.

        Returns:
            Tuple of position vectors of the two frames in the world
                coordinate system.
        """
        pos1_world = configuration.get_transform_frame_to_world(
            self.frames[0]
        ).translation
        pos2_world = configuration.get_transform_frame_to_world(
            self.frames[1]
        ).translation
        return pos1_world, pos2_world

    def _get_frame_jacobians(
        self, configuration: Configuration
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Get the position Jacobians of the two frames in the world frame.

        Args:
            configuration: Robot configuration.

        Returns:
            Tuple of position Jacobian matrices of the two frames in the world
                coordinate system.
        """
        pos1_jac = configuration.get_frame_jacobian(self.frames[0])[:3]
        rotation1 = configuration.get_transform_frame_to_world(
            self.frames[0]
        ).rotation
        pos1_jac = rotation1 @ pos1_jac

        pos2_jac = configuration.get_frame_jacobian(self.frames[1])[:3]
        rotation2 = configuration.get_transform_frame_to_world(
            self.frames[1]
        ).rotation
        pos2_jac = rotation2 @ pos2_jac

        return pos1_jac, pos2_jac



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Kinova Gen2 arm tracking a moving target."""

import meshcat_shapes
import numpy as np
import qpsolvers
from loop_rate_limiters import RateLimiter

import pink
from pink import solve_ik
from pink.tasks import FrameTask, PostureTask
from pink.utils import custom_configuration_vector
from pink.visualization import start_meshcat_visualizer

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try ``pip install robot_descriptions``"
    )


if __name__ == "__main__":
    robot = load_robot_description("gen2_description", root_joint=None)
    viz = start_meshcat_visualizer(robot)

    end_effector_task = FrameTask(
        "j2s6s200_end_effector",
        position_cost=1.0,  # [cost] / [m]
        orientation_cost=1.0,  # [cost] / [rad]
    )

    posture_task = PostureTask(
        cost=1e-3,  # [cost] / [rad]
    )

    tasks = [end_effector_task, posture_task]

    q_ref = custom_configuration_vector(
        robot,
        j2s6s200_joint_2=1.0,
        j2s6s200_joint_3=1.0,
        j2s6s200_joint_5=1.0,
    )
    configuration = pink.Configuration(robot.model, robot.data, q_ref)
    for task in tasks:
        task.set_target_from_configuration(configuration)
    viz.display(configuration.q)

    viewer = viz.viewer
    meshcat_shapes.frame(viewer["end_effector_target"], opacity=0.5)
    meshcat_shapes.frame(viewer["end_effector"], opacity=1.0)

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "quadprog" in qpsolvers.available_solvers:
        solver = "quadprog"

    rate = RateLimiter(frequency=200.0)
    dt = rate.period
    t = 0.0  # [s]
    while True:
        # Update task targets
        end_effector_target = end_effector_task.transform_target_to_world
        end_effector_target.translation[1] = 0.8 + 0.1 * np.sin(2.0 * t)
        end_effector_target.translation[2] = 0.2

        # Update visualization frames
        viewer["end_effector_target"].set_transform(end_effector_target.np)
        viewer["end_effector"].set_transform(
            configuration.get_transform_frame_to_world(
                end_effector_task.frame
            ).np
        )

        # Compute velocity and integrate it into next configuration
        velocity = solve_ik(configuration, tasks, dt, solver=solver)
        configuration.integrate_inplace(velocity, dt)

        # Visualize result at fixed FPS
        viz.display(configuration.q)
        rate.sleep()
        t += dt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Exceptions raised by tasks."""

from ..exceptions import PinkError


class TargetNotSet(PinkError):
    """Exception raised when attempting to compute with an unset target."""


class TaskDefinitionError(PinkError):
    """Exception raised when a task definition is ill-formed."""


class TaskJacobianNotSet(PinkError):
    """Exception raised when attempting to compute without a task Jacobian."""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Test fixture for the linear holonomic task."""

import unittest

import numpy as np
import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.tasks import LinearHolonomicTask, TaskJacobianNotSet
from pink.tasks.exceptions import TaskDefinitionError
from pink.utils import get_joint_idx


class TestLinearHolonomicTask(unittest.TestCase):
    """Test consistency of the linear holonomic task.

    Note:
        This fixture only tests the task itself. Integration tests with the IK
        are carried out in :class:`TestSolveIK`.
    """

    def setUp(self):
        """Prepare test fixture."""
        robot = load_robot_description(
            "draco3_description", root_joint=pin.JointModelFreeFlyer()
        )
        self.configuration = Configuration(robot.model, robot.data, robot.q0)

    def test_inconsistent_definition(self):
        """Exception when the task is not defined properly."""
        with self.assertRaises(TaskDefinitionError):
            LinearHolonomicTask(A=np.ones((3, 4)), b=np.ones(5), q_0=None)

    def test_task_repr(self):
        """String representation reports the task gain, costs and target."""
        A = np.zeros((1, self.configuration.model.nv))
        _, r_knee_fe_jp_v_idx = get_joint_idx(
            self.configuration.model, "r_knee_fe_jp"
        )
        _, r_knee_fe_jd_v_idx = get_joint_idx(
            self.configuration.model, "r_knee_fe_jd"
        )
        A[:, r_knee_fe_jp_v_idx] = 1.0
        A[:, r_knee_fe_jd_v_idx] = -1.0
        task = LinearHolonomicTask(A=A, b=np.zeros(1), q_0=None, cost=1.0)
        self.assertTrue("cost=" in repr(task))
        self.assertTrue("gain=" in repr(task))

    def test_jacobian_not_set_properly(self):
        """Raise an exception when the task Jacobian is not set properly."""
        A = np.zeros((1, self.configuration.model.nq))
        task = LinearHolonomicTask(A=A, b=np.zeros(1), q_0=None, cost=1.0)
        with self.assertRaises(TaskJacobianNotSet):
            task.compute_error(self.configuration)
        with self.assertRaises(TaskJacobianNotSet):
            task.compute_jacobian(self.configuration)

    def test_unit_cost_qp_objective(self):
        """A unit cost vector means the QP objective is (J^T J, -e^T J)."""
        A = np.zeros((2, self.configuration.model.nv))
        _, r_knee_fe_jp_v_idx = get_joint_idx(
            self.configuration.model, "r_knee_fe_jp"
        )
        _, r_knee_fe_jd_v_idx = get_joint_idx(
            self.configuration.model, "r_knee_fe_jd"
        )
        A[:, r_knee_fe_jp_v_idx] = 1.0
        A[:, r_knee_fe_jd_v_idx] = -1.0
        _, l_knee_fe_jp_v_idx = get_joint_idx(
            self.configuration.model, "l_knee_fe_jp"
        )
        _, l_knee_fe_jd_v_idx = get_joint_idx(
            self.configuration.model, "l_knee_fe_jd"
        )
        A[:, l_knee_fe_jp_v_idx] = 1.0
        A[:, l_knee_fe_jd_v_idx] = -1.0
        task = LinearHolonomicTask(
            A=A, b=np.zeros(2), q_0=None, cost=[1.0, 1.0]
        )
        e = task.compute_error(self.configuration)
        J = task.compute_jacobian(self.configuration)
        H, c = task.compute_qp_objective(self.configuration)
        self.assertTrue(np.allclose(J.T @ J, H))
        self.assertTrue(np.allclose(-e.T @ J, c))

    def test_zero_cost_same_as_disabling_task(self):
        """The task has no effect when its cost is zero."""
        A = np.zeros((2, self.configuration.model.nv))
        _, r_knee_fe_jp_v_idx = get_joint_idx(
            self.configuration.model, "r_knee_fe_jp"
        )
        _, r_knee_fe_jd_v_idx = get_joint_idx(
            self.configuration.model, "r_knee_fe_jd"
        )
        A[:, r_knee_fe_jp_v_idx] = 1.0
        A[:, r_knee_fe_jd_v_idx] = -1.0
        _, l_knee_fe_jp_v_idx = get_joint_idx(
            self.configuration.model, "l_knee_fe_jp"
        )
        _, l_knee_fe_jd_v_idx = get_joint_idx(
            self.configuration.model, "l_knee_fe_jd"
        )
        A[:, l_knee_fe_jp_v_idx] = 1.0
        A[:, l_knee_fe_jd_v_idx] = -1.0
        task = LinearHolonomicTask(
            A=A, b=np.zeros(2), q_0=None, cost=[0.0, 0.0]
        )
        J = task.compute_jacobian(self.configuration)
        H, c = task.compute_qp_objective(self.configuration)
        qd = np.random.random(J.shape[1:])
        cost = qd.T @ H @ qd + c @ qd
        self.assertAlmostEqual(cost, 0.0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria

"""Test fixture for other library features."""

import unittest

import os
import numpy as np
import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description
from robot_descriptions.iiwa14_description import PACKAGE_PATH, REPOSITORY_PATH

from pink.exceptions import ConfigurationError
from pink.utils import VectorSpace, custom_configuration_vector, process_collision_pairs


class TestUtils(unittest.TestCase):
    """Test utility classes and functions."""

    def test_custom_configuration_vector(self):
        """Check a custom configuration vector for Upkie.

        Assumes the left and right knees have joint indices respectively 8 and
        11 in the configuration vector.
        """
        robot = load_robot_description("upkie_description", root_joint=pin.JointModelFreeFlyer())
        q = custom_configuration_vector(robot, left_knee=0.2, right_knee=-0.2)
        self.assertAlmostEqual(q[8], 0.2)
        self.assertAlmostEqual(q[11], -0.2)

    def test_custom_configuration_vector_unbounded_joints(self):
        """Single number for an unbounded joint should yield an error."""
        robot = load_robot_description("gen2_description", root_joint=None)
        with self.assertRaises(ConfigurationError):
            custom_configuration_vector(robot, j2s6s200_joint_1=0.0)
        cos = np.sqrt(3) / 2.0
        sin = 0.5
        q = custom_configuration_vector(robot, j2s6s200_joint_1=[cos, sin])
        self.assertAlmostEqual(q[0], cos)
        self.assertAlmostEqual(q[1], sin)

    def test_vector_space(self):
        """Check dimensions of regular tangent space."""
        robot = load_robot_description("upkie_description", root_joint=pin.JointModelFreeFlyer())
        nv = robot.model.nv
        tangent = VectorSpace(robot.model.nv)
        self.assertEqual(tangent.eye.shape, (nv, nv))
        self.assertEqual(tangent.ones.shape, (nv,))
        self.assertEqual(tangent.zeros.shape, (nv,))

    def test_process_collision_pairs(self):
        urdf_path = os.path.join(PACKAGE_PATH, "urdf", "iiwa14_spheres_collision.urdf")
        robot = pin.RobotWrapper.BuildFromURDF(urdf_path, package_dirs=[os.path.dirname(REPOSITORY_PATH)])

        robot.collision_data = process_collision_pairs(robot.model, robot.collision_model)

        # The contacts are enabled
        self.assertTrue(robot.collision_data.enable_contact)
        # The amount of total collision pairs is 72
        self.assertEqual(len(robot.collision_data.distanceResults), 72)

        # If sample srdf file is taken, the number of collision pairs shoul reduce
        robot = pin.RobotWrapper.BuildFromURDF(urdf_path, package_dirs=[os.path.dirname(REPOSITORY_PATH)])
        srdf_path = os.path.dirname(os.path.realpath(__file__)) + "/iiwa_exclude_pairs.srdf"
        robot.collision_data = process_collision_pairs(robot.model, robot.collision_model, srdf_path=srdf_path)
        self.assertTrue(len(robot.collision_data.distanceResults) < 72)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Test inverse kinematics."""

import unittest

import numpy as np
import pinocchio as pin
import qpsolvers
from numpy.linalg import norm
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration, build_ik, solve_ik
from pink.exceptions import NotWithinConfigurationLimits
from pink.tasks import FrameTask, ComTask
from pink.barriers import PositionBarrier


class TestSolveIK(unittest.TestCase):
    """Test fixture for the solve_ik function."""

    def test_checks_configuration_limits(self):
        """IK checks for configuration limits."""
        robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        q = robot.q0
        q[7] = 20  # above limit for Upkie's first joint
        configuration = Configuration(robot.model, robot.data, q)
        with self.assertRaises(NotWithinConfigurationLimits):
            solve_ik(configuration, [], dt=1.0, solver="quadprog")

    def test_ignore_configuration_limits(self):
        """If flag is set, do not check configuration limits."""
        robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        q = robot.q0
        q[7] = 20  # above limit for Upkie's first joint
        configuration = Configuration(robot.model, robot.data, q)

        solve_ik(
            configuration,
            [],
            dt=1.0,
            solver="quadprog",
            safety_break=False,
        )

    def test_model_with_no_joint_limit(self):
        """Model with no joint limit has no inequality constraints."""
        model = pin.Model()
        model.addJoint(
            0, pin.JointModelSpherical(), pin.SE3.Identity(), "spherical"
        )
        robot = pin.RobotWrapper(model=model)
        configuration = Configuration(robot.model, robot.data, robot.q0)
        problem = build_ik(configuration, [], dt=1.0)
        self.assertIsNone(problem.G)
        self.assertIsNone(problem.h)

    def test_no_task(self):
        """Raise an error when the robot body is not found."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        tasks = []
        v = solve_ik(configuration, tasks, dt=1e-3, solver="quadprog")
        self.assertTrue(np.allclose(v, np.zeros(robot.nv)))

    def test_single_task_fulfilled(self):
        """Velocity is zero when the only task is already fulfilled."""
        robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        task = FrameTask(
            "left_contact", position_cost=1.0, orientation_cost=1.0
        )
        task.set_target(
            configuration.get_transform_frame_to_world("left_contact")
        )
        velocity = solve_ik(configuration, [task], dt=5e-3, solver="quadprog")
        self.assertTrue(np.allclose(velocity, 0.0))

    def test_barrier_fullfilled(self):
        """Velocity is zero when the only task is already fulfilled."""
        robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        task = FrameTask(
            "left_contact", position_cost=1.0, orientation_cost=1.0
        )
        barrier = PositionBarrier(
            "left_contact",
            p_min=configuration.get_transform_frame_to_world(
                "left_contact"
            ).translation
            - 0.1 * np.ones(3),
        )
        task.set_target(
            configuration.get_transform_frame_to_world("left_contact")
        )
        velocity = solve_ik(
            configuration,
            [task],
            dt=5e-3,
            solver="quadprog",
            barriers=[barrier],
        )
        self.assertTrue(np.allclose(velocity, 0.0))

    def test_barrier_violated(self):
        """Velocity is zero when the only task is already fulfilled."""
        robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        task = FrameTask(
            "left_contact", position_cost=1.0, orientation_cost=1.0
        )
        barrier = PositionBarrier(
            "left_contact",
            p_min=configuration.get_transform_frame_to_world(
                "left_contact"
            ).translation
            + 1.0 * np.ones(3),
        )
        task.set_target(
            configuration.get_transform_frame_to_world("left_contact")
        )
        velocity = solve_ik(
            configuration,
            [task],
            dt=5e-3,
            solver="quadprog",
            barriers=[barrier],
        )
        self.assertFalse(np.allclose(velocity, 0.0))

    def test_single_task_convergence(self):
        """Integrating velocities makes a task converge to its target."""
        robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        task = FrameTask(
            "left_contact", position_cost=1.0, orientation_cost=1.0
        )
        transform_init_to_world = configuration.get_transform_frame_to_world(
            "left_contact"
        )
        transform_target_to_init = pin.SE3(
            np.eye(3), np.array([0.0, 0.0, 0.1])
        )
        transform_target_to_world = (
            transform_init_to_world * transform_target_to_init
        )
        task.set_target(transform_target_to_world)
        dt = 5e-3  # [s]
        velocity = solve_ik(configuration, [task], dt, solver="quadprog")

        # Initially we are nowhere near the target and moving
        self.assertFalse(np.allclose(velocity, 0.0))
        self.assertAlmostEqual(norm(task.compute_error(configuration)), 0.1)
        self.assertFalse(
            configuration.get_transform_frame_to_world(
                "left_contact"
            ).isApprox(transform_target_to_world, prec=1e-4)
        )

        last_error = 1e6
        for nb_steps in range(42):
            error = norm(task.compute_error(configuration))
            if error < 1e-6 and np.allclose(velocity, 0.0):
                break
            self.assertLess(error, last_error)  # error stictly decreases
            last_error = error
            q = configuration.integrate(velocity, dt)
            configuration = Configuration(robot.model, robot.data, q)
            velocity = solve_ik(configuration, [task], dt, solver="quadprog")

        # After nb_steps we are at the target and not moving
        self.assertTrue(np.allclose(velocity, 0.0))
        self.assertAlmostEqual(norm(task.compute_error(configuration)), 0.0)
        self.assertTrue(
            configuration.get_transform_frame_to_world(
                "left_contact"
            ).isApprox(transform_target_to_world, prec=1e-8)
        )
        self.assertLess(nb_steps, 3)

    def test_single_task_translation(self):
        """Translating a target yields a pure linear velocity."""
        robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        contact_task = FrameTask(
            "right_contact", position_cost=1.0, orientation_cost=1.0
        )
        transform_target_to_world = configuration.get_transform_frame_to_world(
            "right_contact"
        ).copy()
        self.assertTrue(
            np.allclose(transform_target_to_world.rotation, np.eye(3)),
            "default orientation should be aligned with the world frame",
        )
        transform_target_to_world.translation[1] -= 0.1
        contact_task.set_target(transform_target_to_world)
        contact_task.lm_damping = 0.0  # only Tikhonov damping for this test
        velocity = solve_ik(
            configuration,
            [contact_task],
            dt=1e-3,
            damping=1e-12,
            solver="quadprog",
        )
        jacobian_contact_in_contact = configuration.get_frame_jacobian(
            "right_contact"
        )
        velocity_contact_in_contact = jacobian_contact_in_contact @ velocity
        linear_velocity_contact_in_contact = velocity_contact_in_contact[0:3]
        angular_velocity_contact_in_contact = velocity_contact_in_contact[3:6]
        self.assertTrue(np.allclose(angular_velocity_contact_in_contact, 0.0))
        self.assertAlmostEqual(linear_velocity_contact_in_contact[0], 0.0)
        self.assertLess(linear_velocity_contact_in_contact[1], 0.0)
        self.assertAlmostEqual(linear_velocity_contact_in_contact[2], 0.0)

    def test_three_tasks_fulfilled(self):
        """No motion when all targets are reached."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        left_ankle_task = FrameTask(
            "l_ankle", position_cost=1.0, orientation_cost=3.0
        )
        right_ankle_task = FrameTask(
            "r_ankle", position_cost=1.0, orientation_cost=3.0
        )
        pelvis_task = FrameTask(
            "PELVIS_S", position_cost=1.0, orientation_cost=3.0
        )

        left_ankle_task.set_target(
            configuration.get_transform_frame_to_world("l_ankle")
        )
        right_ankle_task.set_target(
            configuration.get_transform_frame_to_world("r_ankle")
        )
        pelvis_task.set_target(
            configuration.get_transform_frame_to_world("PELVIS_S")
        )

        tasks = [pelvis_task, left_ankle_task, right_ankle_task]
        velocity = solve_ik(configuration, tasks, dt=5e-3, solver="quadprog")
        self.assertTrue(np.allclose(velocity, 0.0))

    def test_three_tasks_convergence(self):
        """Three simultaneously feasible tasks on the JVRC model converge."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)

        # Define tasks
        left_ankle_task = FrameTask(
            "l_ankle", position_cost=1.0, orientation_cost=3.0
        )
        right_ankle_task = FrameTask(
            "r_ankle", position_cost=1.0, orientation_cost=3.0
        )
        pelvis_task = FrameTask(
            "PELVIS_S", position_cost=1.0, orientation_cost=0.0
        )
        tasks = [pelvis_task, left_ankle_task, right_ankle_task]

        # Set task targets
        transform_l_ankle_target_to_init = pin.SE3(
            np.eye(3), np.array([0.1, 0.0, 0.0])
        )
        transform_r_ankle_target_to_init = pin.SE3(
            np.eye(3), np.array([-0.1, 0.0, 0.0])
        )
        left_ankle_task.set_target(
            configuration.get_transform_frame_to_world("l_ankle")
            * transform_l_ankle_target_to_init
        )
        right_ankle_task.set_target(
            configuration.get_transform_frame_to_world("r_ankle")
            * transform_r_ankle_target_to_init
        )
        pelvis_task.set_target(
            configuration.get_transform_frame_to_world("PELVIS_S")
        )

        # Run IK in closed loop
        dt = 4e-3  # [s]
        for nb_iter in range(42):
            velocity = solve_ik(configuration, tasks, dt, solver="quadprog")
            if norm(velocity) < 1e-10:
                break
            q = configuration.integrate(velocity, dt)
            configuration = Configuration(robot.model, robot.data, q)
        self.assertLess(nb_iter, 42)
        self.assertLess(norm(velocity), 1e-10)
        self.assertLess(
            max(norm(task.compute_error(configuration)) for task in tasks),
            0.5,
        )

    def test_com_task_fulfilled(self):
        """No motion when all targets are reached."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        left_ankle_task = FrameTask(
            "l_ankle", position_cost=1.0, orientation_cost=3.0
        )
        right_ankle_task = FrameTask(
            "r_ankle", position_cost=1.0, orientation_cost=3.0
        )
        com_task = ComTask(cost=2.0)

        left_ankle_task.set_target(
            configuration.get_transform_frame_to_world("l_ankle")
        )
        right_ankle_task.set_target(
            configuration.get_transform_frame_to_world("r_ankle")
        )
        com_task.set_target_from_configuration(configuration)

        tasks = [com_task, left_ankle_task, right_ankle_task]
        velocity = solve_ik(configuration, tasks, dt=5e-3, solver="quadprog")
        self.assertTrue(np.allclose(velocity, 0.0))

    def test_com_task_convergence(self):
        """Three simultaneously feasible tasks on the JVRC model converge."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)

        # Define tasks
        left_ankle_task = FrameTask(
            "l_ankle", position_cost=1.0, orientation_cost=3.0
        )
        right_ankle_task = FrameTask(
            "r_ankle", position_cost=1.0, orientation_cost=3.0
        )
        com_task = ComTask(cost=2.0)
        tasks = [com_task, left_ankle_task, right_ankle_task]

        # Set task targets
        transform_l_ankle_target_to_init = pin.SE3(
            np.eye(3), np.array([0.1, 0.0, 0.0])
        )
        transform_r_ankle_target_to_init = pin.SE3(
            np.eye(3), np.array([-0.1, 0.0, 0.0])
        )
        left_ankle_task.set_target(
            configuration.get_transform_frame_to_world("l_ankle")
            * transform_l_ankle_target_to_init
        )
        right_ankle_task.set_target(
            configuration.get_transform_frame_to_world("r_ankle")
            * transform_r_ankle_target_to_init
        )

        # Set the desired CoM target 0.05 above the initial CoM position
        initial_com = pin.centerOfMass(
            robot.model, robot.data, configuration.q
        )
        desired_com = initial_com.copy()
        desired_com[2] += 0.05
        com_task.set_target(desired_com)

        # Run IK in closed loop
        dt = 4e-3  # [s]
        for nb_iter in range(42):
            velocity = solve_ik(configuration, tasks, dt, solver="quadprog")
            if np.linalg.norm(velocity) < 1e-10:
                break
            configuration.integrate_inplace(velocity, dt)

        self.assertLess(nb_iter, 42)
        self.assertLess(np.linalg.norm(velocity), 1e-10)
        self.assertLess(
            max(
                np.linalg.norm(task.compute_error(configuration))
                for task in tasks
            ),
            0.5,
        )

    def get_jvrc_problem(self):
        """Get an IK problem with three tasks on a humanoid model."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        left_ankle_task = FrameTask(
            "l_ankle", position_cost=1.0, orientation_cost=3.0
        )
        right_ankle_task = FrameTask(
            "r_ankle", position_cost=1.0, orientation_cost=3.0
        )
        pelvis_task = FrameTask(
            "PELVIS_S", position_cost=1.0, orientation_cost=3.0
        )
        left_ankle_task.set_target(
            configuration.get_transform_frame_to_world("l_ankle")
        )
        right_ankle_task.set_target(
            configuration.get_transform_frame_to_world("r_ankle")
        )
        pelvis_task.set_target(
            configuration.get_transform_frame_to_world("PELVIS_S")
        )
        tasks = [pelvis_task, left_ankle_task, right_ankle_task]
        dt = 5e-3
        return configuration, tasks, dt

    @staticmethod
    def get_solver_test(solver: str):
        """Get test function for a given QP solver.

        Args:
            solver: Name of the QP solver to use as backend.

        Returns:
            Test function for that solver.
        """

        def test(self):
            configuration, tasks, dt = self.get_jvrc_problem()
            solve_ik(configuration, tasks, dt, solver=solver)

        return test


# Generate test fixtures for each solver
for solver in qpsolvers.available_solvers:
    if solver != "highs":
        setattr(
            TestSolveIK, f"test_{solver}", TestSolveIK.get_solver_test(solver)
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Posture task implementation."""

from typing import Optional

import numpy as np
import pinocchio as pin

from ..configuration import Configuration
from ..utils import get_root_joint_dim
from .exceptions import TargetNotSet
from .task import Task


class PostureTask(Task):
    r"""Regulate joint angles to a desired posture.

    A posture is a vector of actuated joint angles. Floating base coordinates
    are not affected by this task.

    Attributes:
        target_q: Target vector in the configuration space.

    A posture task is typically used for regularization as it has a steady
    rank. For instance, when Upkie's legs are stretched and the Jacobian of its
    contact frames become singular, the posture task will drive the knees
    toward a preferred orientation.
    """

    target_q: Optional[np.ndarray]

    def __init__(
        self, cost: float, lm_damping: float = 0.0, gain: float = 1.0
    ) -> None:
        r"""Create task.

        Args:
            cost: value used to cast joint angle differences to a homogeneous
                cost, in :math:`[\mathrm{cost}] / [\mathrm{rad}]`.
            lm_damping: Unitless scale of the Levenberg-Marquardt (only when
                the error is large) regularization term, which helps when
                targets are unfeasible. Increase this value if the task is too
                jerky under unfeasible targets, but beware that too large a
                damping can slow down the task.
            gain: Task gain :math:`\alpha \in [0, 1]` for additional low-pass
                filtering. Defaults to 1.0 (no filtering) for dead-beat
                control.

        Note:
            We assume that the first seven coordinates of the configuration are
            for the floating base.
        """
        super().__init__(cost=cost, gain=gain, lm_damping=lm_damping)
        self.target_q = None

    def set_target(self, target_q: np.ndarray) -> None:
        """Set target posture.

        Args:
            target_q: Target vector in the configuration space.
        """
        self.target_q = target_q.copy()

    def set_target_from_configuration(
        self, configuration: Configuration
    ) -> None:
        """Set target posture from a robot configuration.

        Args:
            configuration: Robot configuration.
        """
        self.set_target(configuration.q)

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute posture task error.

        The posture task error is defined as:

        .. math::

            e(q) = q^* \ominus q

        See :func:`pink.tasks.task.Task.compute_error` for more context.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Posture task error :math:`e(q)`.
        """
        if self.target_q is None:
            raise TargetNotSet("no posture target")
        _, nv = get_root_joint_dim(configuration.model)
        return pin.difference(
            configuration.model,
            self.target_q,
            configuration.q,
        )[nv:]

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the posture task Jacobian.

        The task Jacobian is the identity :math:`I_{n_v} \in \mathbb{R}^{n_v
        \times n_v}`, with :math:`n_v` the dimension of the robot's tangent
        space, so that the task dynamics are:

        .. math::

            J(q) \Delta q = \Delta q = \alpha (q^* \ominus q)

        See :func:`pink.tasks.task.Task.compute_jacobian` for more context.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Posture task Jacobian :math:`J(q)`.
        """
        _, nv = get_root_joint_dim(configuration.model)
        return configuration.tangent.eye[nv:, :]

    def __repr__(self):
        """Human-readable representation of the task."""
        return (
            "PostureTask("
            f"cost={self.cost}, "
            f"gain={self.gain}, "
            f"lm_damping={self.lm_damping})"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Configuration space of a robot model.

Pink uses `Pinocchio <https://github.com/stack-of-tasks/pinocchio>`__ for
forward kinematics. A :class:`Configuration` is a pair of Pinocchio model and
data where forward kinematics have been run, indicating that frame transforms
and frame Jacobians used for IK can be queried.
"""

import logging
from typing import Optional

import numpy as np
import pinocchio as pin

from .exceptions import FrameNotFound, NotWithinConfigurationLimits
from .limits import ConfigurationLimit, VelocityLimit
from .utils import VectorSpace, get_root_joint_dim


class Configuration:
    """Type indicating that configuration-dependent quantities are available.

    In Pink, this type enables access to frame transforms and frame Jacobians.
    We rely on typing to make sure the proper forward kinematics functions have
    been called beforehand. In Pinocchio, these functions are:

    - ``pin.computeJointJacobians(model, data, configuration)``
    - ``pin.updateFramePlacements(model, data)``

    The former computes the full model Jacobian into ``data.J``. (It also
    computes forward kinematics, so there is no need to further call
    ``pin.forwardKinematics(model, data, configuration)``.) The latter updates
    frame placements.

    Additionally, if collision model is provided, it is used to evaluate
    distances between frames using following functions:

    - ``pin.computeCollisions(model, data, collision_model, collision_data, q)`` # noqa: E501
    - ``pin.updateGeometryPlacements(model, data, collision_model, collision_data, q)``  # noqa: E501

    Notes:
        This class is meant to be used as a subclass of pin.RobotWrapper, not
        wrap it. However, right now pin.RobotWrapper does not have a shallow
        copy constructor. TODO(scaron): bring it up upstream.

    Attributes:
        data: Data corresponding to :data:`Configuration.model`.
        model: Kinodynamic model.
        collision_data: Data corresponding to :data:`Configuration.collision_model`.
        collision_model: Collision model.
        q: Configuration vector for the robot model.
    """

    collision_data: pin.GeometryData
    collision_model: pin.GeometryModel
    data: pin.Data
    model: pin.Model
    q: np.ndarray

    def __init__(
        self,
        model: pin.Model,
        data: pin.Data,
        q: np.ndarray,
        copy_data: bool = True,
        forward_kinematics: bool = True,
        collision_model: Optional[pin.GeometryModel] = None,
        collision_data: Optional[pin.GeometryData] = None,
    ):
        """Initialize configuration.

        Args:
            model: Kinodynamic model.
            data: Data corresponding to the model.
            q: Configuration vector.
            copy_data: If true (default), work on an internal copy of the input
                data. Otherwise, work on the input data directly.
            forward_kinematics: If true (default), compute forward kinematics
                from the q into the internal data.
            collision_model: collision geometry model, with already loaded
                collisions. Default is None, meaning no collisions are
                processeed.
            collision_data: collision geometry data, with already loaded
                collisions. Default is None, which generates it from the model
                if possible, None otherwise.

        Notes:
            Configurations copy data and run forward kinematics by default so
            that they are less error-prone for newcomers. You can avoid copies
            or forward kinematics (e.g. if it is already computed by the
            caller) using constructor parameters.
        """
        if not hasattr(model, "tangent"):
            model.tangent = VectorSpace(model.nv)
        if not hasattr(model, "configuration_limit"):
            model.configuration_limit = ConfigurationLimit(model)
        if not hasattr(model, "velocity_limit"):
            model.velocity_limit = VelocityLimit(model)
        q_readonly = q.copy()
        q_readonly.setflags(write=False)
        self.data = data.copy() if copy_data else data
        self.model = model
        self.q = q_readonly
        self.tangent = model.tangent

        # Update collision
        self.collision_model = collision_model
        self.collision_data = (
            collision_data
            if collision_data is not None
            else (
                pin.GeometryData(collision_model)
                if collision_model is not None
                else None
            )
        )

        if forward_kinematics:
            self.update(None)

    def update(self, q: Optional[np.ndarray] = None) -> None:
        """Update configuration to a new vector and run forward kinematics and
        collision pairs distance calculations, if specified.

        Args:
            q: New configuration vector.
        """
        if q is not None:
            q_readonly = q.copy()
            q_readonly.setflags(write=False)
            self.q = q_readonly

        # Compute collisions, if needed
        if self.collision_model is not None:
            pin.computeCollisions(
                self.model,
                self.data,
                self.collision_model,
                self.collision_data,
                self.q,
                False,
            )
            pin.computeDistances(
                self.model,
                self.data,
                self.collision_model,
                self.collision_data,
                self.q,
            )

        pin.computeJointJacobians(self.model, self.data, self.q)
        pin.updateFramePlacements(self.model, self.data)

    def check_limits(
        self, tol: float = 1e-6, safety_break: bool = True
    ) -> None:
        """Check that the current configuration is within limits.

        Args:
            tol: Tolerance in radians.
            safe_break (bool): If True, stop execution and raise an exception
                if the current configuration is outside limits. If False, print
                a warning and continue execution.

        Raises:
            NotWithinConfigurationLimits: If the current configuration is
                outside limits.
        """
        q_max = self.model.upperPositionLimit
        q_min = self.model.lowerPositionLimit
        root_nq, _ = get_root_joint_dim(self.model)
        for i in range(root_nq, self.model.nq):
            if q_max[i] <= q_min[i] + tol:  # no limit
                continue
            if self.q[i] < q_min[i] - tol or self.q[i] > q_max[i] + tol:
                if safety_break:
                    raise NotWithinConfigurationLimits(
                        i,
                        self.q[i],
                        q_min[i],
                        q_max[i],
                    )
                else:
                    logging.warning(
                        "Value %f at index %d is out of limits: [%f, %f]",
                        self.q[i],
                        i,
                        q_min[i],
                        q_max[i],
                    )

    def get_frame_jacobian(self, frame: str) -> np.ndarray:
        r"""Compute the Jacobian matrix of a frame velocity.

        Denoting our frame by :math:`B` and the world frame by :math:`W`, the
        Jacobian matrix :math:`{}_B J_{WB}` is related to the body velocity
        :math:`{}_B v_{WB}` by:

        .. math::

            {}_B v_{WB} = {}_B J_{WB} \dot{q}

        Args:
            frame: Name of the frame, typically a link name from the URDF.

        Returns:
            Jacobian :math:`{}_B J_{WB}` of the frame.

        When the robot model includes a floating base
        (pin.JointModelFreeFlyer), the configuration vector :math:`q` consists
        of:

        - ``q[0:3]``: position in [m] of the floating base in the inertial
          frame, formatted as :math:`[p_x, p_y, p_z]`.
        - ``q[3:7]``: unit quaternion for the orientation of the floating base
          in the inertial frame, formatted as :math:`[q_x, q_y, q_z, q_w]`.
        - ``q[7:]``: joint angles in [rad].
        """
        if not self.model.existFrame(frame):
            raise FrameNotFound(frame, self.model.frames)
        frame_id = self.model.getFrameId(frame)
        J: np.ndarray = pin.getFrameJacobian(
            self.model, self.data, frame_id, pin.ReferenceFrame.LOCAL
        )
        return J

    def get_transform_frame_to_world(self, frame: str) -> pin.SE3:
        """Get the pose of a frame in the current configuration.

        Args:
            frame: Name of a frame, typically a link name from the URDF.

        Returns:
            Current transform from the given frame to the world frame.

        Raises:
            KeyError: if the frame name is not found in the robot model.
        """
        frame_id = self.model.getFrameId(frame)
        try:
            return self.data.oMf[frame_id].copy()
        except IndexError as index_error:
            raise FrameNotFound(frame, self.model.frames) from index_error

    def get_transform(self, source: str, dest: str) -> pin.SE3:
        """Get the pose of a frame with respect to another frame.

        Args:
            source: Name of the frame to get the pose of.
            dest: Name of the frame to get the pose in.

        Returns:
            Current transform from the source frame to the dest frame.

        Raises:
            KeyError: if any of the frame names is not found in the model.
        """
        transform_source_to_world = self.get_transform_frame_to_world(source)
        transform_dest_to_world = self.get_transform_frame_to_world(dest)
        return transform_dest_to_world.actInv(transform_source_to_world)

    def integrate(self, velocity, dt) -> np.ndarray:
        """Integrate a velocity starting from the current configuration.

        Args:
            velocity: Velocity in tangent space.
            dt: Integration duration in [s].

        Returns:
            New configuration vector after integration.
        """
        return pin.integrate(self.model, self.q, velocity * dt)

    def integrate_inplace(self, velocity, dt) -> None:
        """Integrate a velocity starting from the current configuration.

        Args:
            velocity: Velocity in tangent space.
            dt: Integration duration in [s].
        """
        q = pin.integrate(self.model, self.q, velocity * dt)
        self.update(q)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Ivan Domrachev, Simeon Nedelchev

"""Frame position barrier."""

from typing import List, Optional, Union

import numpy as np

from ..configuration import Configuration
from .barrier import Barrier
from .exceptions import NoPositionLimitProvided


class PositionBarrier(Barrier):
    r"""A position-based barrier.

    Defines a barrier function based on the position of a
    specified frame in the world coordinate system. It allows for the
    specification of minimum and maximum position bounds along selected axes.

    Attributes:
        frame: Name of the frame to monitor.
        indices: Indices of the position components to consider.
        p_min: Minimum position bounds.
        p_max: Maximum position bounds.
    """

    frame: str
    indices: List[int]
    p_min: Optional[np.ndarray]
    p_max: Optional[np.ndarray]

    def __init__(
        self,
        frame: str,
        indices: Optional[List[int]] = None,
        p_min: Optional[np.ndarray] = None,
        p_max: Optional[np.ndarray] = None,
        gain: Union[float, np.ndarray] = 1.0,
        safe_displacement_gain: float = 0.0,
    ):
        """Initialize a position barrier.

        Args:
            frame: name of the frame to monitor.
            indices: indices with limits.
                Indices from 0 to 2 stand for coordinates x to z respectively.
                Defaults to all dimensions.
            p_min: minimum position limit.
                Defaults to None, meaning no minimum limit is applied.
            p_max: maximum position limit.
                Defaults to None, meaning no maximum limit is applied.
            gain: regularization gain.
                Defaults to 1.0.
            safe_displacement_gain: gain for the safe backup displacement
                cost term. Defaults to 0.0.

        Raises:
            NoPositionLimitProvided: neither minimum nor maximum
                position limit is provided.
        """
        indices = [0, 1, 2] if indices is None else indices

        if p_min is None and p_max is None:
            raise NoPositionLimitProvided(
                f"Position barrier for frame {frame} requires either \
                    p_min or p_max"
            )

        dim = 0
        if p_min is not None:
            dim += len(indices)
        if p_max is not None:
            dim += len(indices)

        # If both p_min and p_max are specified, they must have
        # the same gains repeated along the dimension
        if isinstance(gain, np.ndarray) and len(gain) != dim:
            gain = np.tile(gain, 2)

        super().__init__(
            dim,
            gain=gain,
            safe_displacement_gain=safe_displacement_gain,
        )

        self.indices = indices
        self.frame = frame
        self.p_min = p_min
        self.p_max = p_max

    def compute_barrier(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the value of the barrier function.

        The barrier function is computed based on the position of the specified
        frame in the world coordinate system. It considers the minimum and
        maximum position bounds along the selected axes.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Value of the barrier function
            :math:`h(q)`.
        """
        pos_world = configuration.get_transform_frame_to_world(
            self.frame
        ).translation
        barriers = []
        if self.p_min is not None:
            barriers.append(pos_world[self.indices] - self.p_min)
        if self.p_max is not None:
            barriers.append(self.p_max - pos_world[self.indices])

        return np.concatenate(barriers)

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the Jacobian matrix of the barrier function.

        The Jacobian matrix is computed based on the position Jacobian of the
        specified frame. The Jacobian is transformed to align with the world
        coordinate system and only the selected indices are considered.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Jacobian matrix
            :math:`\frac{\partial h}{\partial q}(q)`.
        """
        pos_jacobian_in_frame = configuration.get_frame_jacobian(self.frame)[
            :3
        ]
        # Transform jacobian to world aligned frame
        rotation = configuration.get_transform_frame_to_world(
            self.frame
        ).rotation
        # Apply rotation to trasfrom from world aligned to local frame
        # and select only relevant indices
        pos_jacobian_in_world = (rotation @ pos_jacobian_in_frame)[
            self.indices
        ]

        jacobians = []
        if self.p_min is not None:
            jacobians.append(pos_jacobian_in_world.copy())
        if self.p_max is not None:
            jacobians.append(-pos_jacobian_in_world.copy())

        return np.vstack(jacobians)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Test base class for kinematic tasks."""

import unittest

from pink.tasks import Task


class TestTask(unittest.TestCase):
    """Test abstract base class for tasks."""

    def setUp(self):
        """Prepare test fixture."""
        Task.__abstractmethods__ = set()  # allow instantiation

    def test_task_repr(self):
        """String representation reports the task gain."""
        task = Task()
        self.assertTrue("gain=" in repr(task))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""Test acceleration limit."""

import unittest

import numpy as np
import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration, solve_ik
from pink.limits import AccelerationLimit
from pink.tasks import FrameTask


class TestAccelerationLimit(unittest.TestCase):
    """Test fixture for acceleration limit."""

    def setUp(self):
        """Set test fixture up."""
        robot = load_robot_description("ur3_description")
        model = robot.model
        a_max = np.ones(robot.model.nv)
        self.a_max = a_max
        self.data = robot.data
        self.q_ref = robot.q0
        self.limit = AccelerationLimit(model, a_max)
        self.model = model

    def test_dimensions(self):
        """Check dimensions of configuration limit projection."""
        nb = len(self.limit.indices)
        nv = self.model.nv
        self.assertEqual(self.limit.projection_matrix.shape, (nb, nv))

    def test_model_with_no_limit(self):
        """Check that unbounded models don't fail."""
        empty_model = pin.Model()
        empty_bounded = AccelerationLimit(empty_model, self.a_max)
        self.assertEqual(len(empty_bounded.indices), 0)
        self.assertIsNone(
            empty_bounded.compute_qp_inequalities(np.empty(0), 1e-3)
        )

    def test_limit_has_an_effect(self):
        """Check that the limit has an effect on a reaching task."""
        end_effector_task = FrameTask(
            "ee_link",
            position_cost=1.0,  # [cost] / [m]
            orientation_cost=1.0,  # [cost] / [rad]
        )
        configuration = Configuration(self.model, self.data, self.q_ref)
        end_effector_task.set_target_from_configuration(configuration)
        end_effector_target = end_effector_task.transform_target_to_world
        end_effector_target.translation[1] = 0.3
        end_effector_target.translation[2] = 0.2
        configuration_limit = configuration.model.configuration_limit
        velocity_limit = configuration.model.velocity_limit
        tasks = [end_effector_task]
        dt = 5e-3
        v_prev = solve_ik(
            configuration,
            tasks,
            dt,
            solver="quadprog",
            limits=[configuration_limit, velocity_limit],
        )
        configuration.integrate_inplace(v_prev, dt)
        self.limit.set_last_integration(v_prev, dt)
        end_effector_target.translation[1] = 0.6
        v_with = solve_ik(
            configuration,
            tasks,
            dt,
            solver="quadprog",
            limits=[configuration_limit, velocity_limit, self.limit],
        )
        v_without = solve_ik(
            configuration,
            tasks,
            dt,
            solver="quadprog",
            limits=[configuration_limit, velocity_limit],
        )
        a_with = (v_with - v_prev) / dt
        a_without = (v_without - v_prev) / dt
        tolerance = 1e-10  # [rad] / [s]^2
        self.assertTrue(np.all(np.abs(a_with) < self.a_max + tolerance))
        self.assertFalse(np.all(np.abs(a_without) < self.a_max + tolerance))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Tests that should pass for all limits."""

import unittest

import numpy as np
import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.limits import ConfigurationLimit, VelocityLimit
from pink.solve_ik import build_ik


class TestLimits(unittest.TestCase):
    """Tests that should pass for all limits."""

    def test_limit_dimension(self):
        """Velocity limit dimension is the number of bounded joints."""
        robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        model = robot.model
        dt = 1e-3  # [s]
        for limit in [ConfigurationLimit(model), VelocityLimit(model)]:
            G, h = limit.compute_qp_inequalities(robot.q0, dt)
            self.assertEqual(G.shape[0], h.shape[0])
            self.assertEqual(G.shape[1], model.nv)

    def test_model_with_no_joint_limit(self):
        """Model with no joint limit has no velocity-limit vector."""
        model = pin.Model()
        model.addJoint(
            0, pin.JointModelSpherical(), pin.SE3.Identity(), "spherical"
        )
        robot = pin.RobotWrapper(model=model)
        dt = 1e-3  # [s]
        for limit in [ConfigurationLimit(model), VelocityLimit(model)]:
            return_value = limit.compute_qp_inequalities(robot.q0, dt)
            self.assertIsNone(return_value)

    def test_model_with_limitless_joint(self):
        """Same as previous test, but the joint has a limit set to zero."""
        model = pin.Model()
        model.addJoint(
            0,
            pin.JointModelRevoluteUnaligned(),
            pin.SE3.Identity(),
            "revolute",
            max_effort=np.array([0.0]),
            max_velocity=np.array([0.0]),
            min_config=np.array([0.0]),
            max_config=np.array([0.0]),
        )
        robot = pin.RobotWrapper(model=model)
        dt = 1e-3  # [s]
        for limit in [ConfigurationLimit(model), VelocityLimit(model)]:
            return_value = limit.compute_qp_inequalities(robot.q0, dt)
            self.assertIsNone(return_value)

    def test_velocity_without_configuration_limits(self, tol: float = 1e-10):
        """Velocity limits are loaded for a model without config limits.

        Args:
            tol: Equality test tolerance.

        Note:
            The loaded robot description is commit-pinned to make sure we load
            a URDF without configuration limits.
        """
        dt = 1e-3  # [s]
        sigmaban = load_robot_description(
            "sigmaban_description",
            commit="d5d023fd35800d00d7647000bce8602617a4960d",
        )
        configuration = Configuration(
            sigmaban.model, sigmaban.data, sigmaban.q0
        )
        problem = build_ik(configuration, [], dt)
        G, h = problem.G, problem.h
        self.assertIsNotNone(G)
        self.assertIsNotNone(h)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Test the Configuration type."""

import unittest

import numpy as np
import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.exceptions import FrameNotFound, NotWithinConfigurationLimits


class TestConfiguration(unittest.TestCase):
    """Test configuration type."""

    def test_constructor(self):
        """Constructing a configuration computes Jacobians."""
        robot = load_robot_description(
            "jvrc_description",
            root_joint=pin.JointModelFreeFlyer(),
            commit="9ff8efbc7043459a8f0892662bd030d8020fb682",
        )
        robot.data.J.fill(42.0)
        configuration = Configuration(robot.model, robot.data, robot.q0)
        J_check = np.array(
            [
                [
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.096,
                    0.389,
                    0.0,
                    0.746,
                    -0.0,
                    -0.0,
                    -0.096,
                    0.389,
                    0.0,
                    0.746,
                    0.0,
                    -0.192,
                    0.0,
                    -0.522,
                    0.0,
                    0.24,
                    -0.217,
                    0.24,
                    0.0,
                    0.24,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    -0.645,
                    -0.522,
                    -0.0,
                    -0.24,
                    -0.217,
                    -0.24,
                    0.0,
                    -0.24,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    -0.746,
                    -0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    -0.746,
                    -0.0,
                    0.0,
                    0.0,
                    0.192,
                    0.0,
                    0.522,
                    0.0,
                    0.0,
                    -0.0,
                    -0.022,
                    -0.0,
                    -0.106,
                    -0.151,
                    -0.106,
                    -0.151,
                    -0.106,
                    -0.151,
                    0.003,
                    0.645,
                    0.0,
                    0.0,
                    0.522,
                    0.0,
                    0.0,
                    -0.0,
                    -0.022,
                    -0.0,
                    -0.106,
                    -0.151,
                    -0.106,
                    -0.151,
                    -0.106,
                    -0.151,
                ],
                [
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    -0.096,
                    0.0,
                    -0.02,
                    -0.096,
                    0.02,
                    0.0,
                    0.096,
                    0.0,
                    -0.02,
                    0.096,
                    0.02,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    -0.24,
                    0.0,
                    0.004,
                    0.0,
                    -0.24,
                    0.0,
                    -0.256,
                    -0.256,
                    -0.256,
                    -0.256,
                    -0.207,
                    -0.207,
                    -0.0,
                    -0.0,
                    -0.003,
                    0.0,
                    0.24,
                    0.0,
                    0.004,
                    0.0,
                    0.24,
                    0.0,
                    0.256,
                    0.256,
                    0.256,
                    0.256,
                    0.207,
                    0.207,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
            ]
        )
        self.assertTrue(np.allclose(configuration.data.J, J_check))

    def test_copy_no_forward_kinematics(self):
        """Refer to the input data, compute forward kinematics."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        robot.data.J.fill(42.0)
        configuration = Configuration(
            robot.model,
            robot.data,
            robot.q0,
            copy_data=False,
            forward_kinematics=True,
        )
        self.assertFalse(np.allclose(configuration.data.J, 42.0))
        configuration.data.J.fill(12.0)
        self.assertTrue(np.allclose(robot.data.J, 12.0))

    def test_no_copy_no_forward_kinematics(self):
        """Not copying means working directly on the underlying data."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        robot.data.J.fill(42.0)
        configuration = Configuration(
            robot.model,
            robot.data,
            robot.q0,
            copy_data=False,
            forward_kinematics=False,
        )
        self.assertTrue(np.allclose(configuration.data.J, 42.0))
        configuration.data.J.fill(12.0)
        self.assertTrue(np.allclose(robot.data.J, 12.0))

    def test_transform_found(self):
        """Return the pose of an existing robot body."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        transform_pelvis_to_world = configuration.get_transform_frame_to_world(
            "PELVIS_S"
        )
        self.assertTrue(
            np.allclose(
                transform_pelvis_to_world.np[3, :],
                np.array([0.0, 0.0, 0.0, 1.0]),
            )
        )

    def test_transform_not_found(self):
        """Raise an error when the request robot body is not found."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        with self.assertRaises(FrameNotFound):
            configuration.get_transform_frame_to_world("foo")

    def test_check_limits(self):
        """Raise an error if and only if a joint limit is exceened."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        q = robot.q0
        configuration = Configuration(robot.model, robot.data, q)
        configuration.check_limits()
        q[-10] += 1e4  # move configuration out of bounds
        configuration = Configuration(robot.model, robot.data, q)
        with self.assertRaises(NotWithinConfigurationLimits):
            configuration.check_limits()

    def test_q_is_a_read_only_copy(self):
        """The `q` attribute of a configuration is a read-only copy."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        original_q = robot.q0
        configuration = Configuration(robot.model, robot.data, original_q)
        the_answer = 42.0
        self.assertNotEqual(configuration.q[2], the_answer)
        original_q[2] = the_answer
        self.assertNotEqual(configuration.q[2], the_answer)
        with self.assertRaises(ValueError):
            configuration.q[2] += 3.0  # read-only

    def test_tangent_eye(self):
        """Configuration's tangent eye is an identity matrix."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        v = np.array([i for i in range(robot.model.nv)])
        self.assertTrue(np.allclose(configuration.tangent.eye.dot(v), v))

    def test_tangent_ones(self):
        """Configuration's tangent ones is a vector of 1.0's."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        self.assertEqual(np.sum(configuration.tangent.ones), robot.model.nv)

    def test_tangent_zeros(self):
        """Configuration's tangent ones is a vector of 0.0's."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        self.assertAlmostEqual(np.sum(configuration.tangent.zeros), 0.0)
        self.assertEqual(len(configuration.tangent.zeros), robot.model.nv)

    def test_frame_jacobian_not_found(self):
        """Querying a body that does not exist raises a ValueError."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        with self.assertRaises(FrameNotFound):
            configuration.get_frame_jacobian("does_not_exist")

    def test_get_integrate_inplace(self):
        """Test in-place integration."""
        robot = load_robot_description("sigmaban_description", root_joint=None)
        configuration = Configuration(robot.model, robot.data, robot.q0)
        velocity = robot.model.tangent.ones
        configuration.integrate_inplace(velocity, dt=1e-3)
        self.assertGreater(np.linalg.norm(configuration.q - robot.q0), 2e-3)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Swing the double pendulum left and right."""

import os

import meshcat_shapes
import numpy as np
import pinocchio as pin
import qpsolvers
from loop_rate_limiters import RateLimiter

import pink
from pink import solve_ik
from pink.tasks import FrameTask, PostureTask
from pink.visualization import start_meshcat_visualizer

if __name__ == "__main__":
    # Load robot description
    urdf_path = os.path.join(
        os.path.dirname(__file__),
        "robots",
        "double_pendulum.urdf",
    )
    robot = pin.RobotWrapper.BuildFromURDF(
        filename=urdf_path,
        package_dirs=["."],
        root_joint=None,
    )

    # Initialize visualizer
    viz = start_meshcat_visualizer(robot)
    viewer = viz.viewer
    meshcat_shapes.frame(viewer["target_frame"], opacity=0.5)
    meshcat_shapes.frame(viewer["tip_frame"], opacity=1.0)

    # Define tasks
    tasks = {
        "tip": FrameTask(
            "link3",
            position_cost=1.0,  # [cost] / [m]
            orientation_cost=1e-3,  # [cost] / [rad]
        ),
        "posture": PostureTask(
            cost=1e-2,  # [cost] / [rad]
        ),
    }

    # Initialize tasks from the initial configuration
    configuration = pink.Configuration(robot.model, robot.data, robot.q0)
    for task in tasks.values():
        task.set_target_from_configuration(configuration)
    viz.display(configuration.q)

    # Homework: what happens if we replace -= by += in the following line?
    tasks["tip"].transform_target_to_world.translation[2] -= 0.1

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "quadprog" in qpsolvers.available_solvers:
        solver = "quadprog"

    rate = RateLimiter(frequency=100.0)
    dt = rate.period
    t = 0.0  # [s]
    while True:
        # Update task targets
        T = tasks["tip"].transform_target_to_world
        T.translation[1] = 0.1 * np.sin(t)

        # Update visualizer frames
        viewer["target_frame"].set_transform(T.np)
        viewer["tip_frame"].set_transform(
            configuration.get_transform_frame_to_world(tasks["tip"].frame).np
        )

        # Compute velocity and integrate it into next configuration
        velocity = solve_ik(configuration, tasks.values(), dt, solver=solver)
        configuration.integrate_inplace(velocity, dt)

        # Visualize result at fixed FPS
        viz.display(configuration.q)
        rate.sleep()
        t += dt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria

"""Subset of bounded joints associated with a robot model."""

from typing import List, Optional, Tuple

import numpy as np
import pinocchio as pin

from .limit import Limit


class ConfigurationLimit(Limit):
    """Subspace of the tangent space restricted to joints with position limits.

    Attributes:
        config_limit_gain: gain between 0 and 1 to steer away from
            configuration limits. It is described in "Real-time prioritized
            kinematic control under inequality constraints for redundant
            manipulators" (Kanoun, 2012). More details in `this writeup
            <https://scaron.info/teaching/inverse-kinematics.html>`__.
        model: Robot model the limit applies to.
        joints: Joints with configuration limits.
        projection_matrix: Projection from tangent space to subspace with
            configuration-limited joints.
    """

    config_limit_gain: float
    indices: np.ndarray
    model: pin.Model
    joints: list
    projection_matrix: Optional[np.ndarray]

    def __init__(self, model: pin.Model, config_limit_gain: float = 0.5):
        """Initialize bounded tangent of a model.

        Args:
            model: robot model.
            config_limit_gain: gain between 0 and 1 to steer away from
                configuration limits. See the full class documentation.
        """
        assert 0.0 < config_limit_gain <= 1.0

        has_configuration_limit = np.logical_and(
            model.hasConfigurationLimit(),
            np.logical_and(
                model.upperPositionLimit < 1e20,
                model.upperPositionLimit > model.lowerPositionLimit + 1e-10,
            ),
        )

        joints = [
            joint
            for joint in model.joints
            if joint.idx_q >= 0
            and has_configuration_limit[
                slice(joint.idx_q, joint.idx_q + joint.nq)
            ].all()
        ]

        index_list: List[int] = []
        for joint in joints:
            index_list.extend(range(joint.idx_v, joint.idx_v + joint.nv))
        indices = np.array(index_list)
        indices.setflags(write=False)

        dim = len(indices)
        projection_matrix = np.eye(model.nv)[indices] if dim > 0 else None

        self.config_limit_gain = config_limit_gain
        self.indices = indices
        self.joints = joints
        self.model = model
        self.projection_matrix = projection_matrix

    def compute_qp_inequalities(
        self,
        q: np.ndarray,
        dt: float,
    ) -> Optional[Tuple[np.ndarray, np.ndarray]]:
        r"""Compute the configuration-dependent velocity limits.

        Those limits are returned as:

        .. math::

            {q \ominus q_{min}} \leq \Delta q \leq {q_{max} \ominus q}

        where :math:`q \in {\cal C}` is the robot's configuration and
        :math:`\Delta q \in T_q({\cal C})` is the displacement in the tangent
        space at :math:`q`. These limits correspond to the derivative of
        :math:`q_{min} \leq q \leq q_{max}`.

        Args:
            q: Robot configuration.
            dt: Integration timestep in [s].

        Returns:
            Pair :math:`(G, h)` representing the inequality constraint as
            :math:`G \Delta q \leq h`, or ``None`` if there is no limit.
        """
        if self.projection_matrix is None:  # no joint (thus checked for mypy)
            return None

        Delta_q_max = pin.difference(
            self.model, q, self.model.upperPositionLimit
        )
        Delta_q_min = pin.difference(
            self.model, q, self.model.lowerPositionLimit
        )
        p_max = self.config_limit_gain * Delta_q_max[self.indices]
        p_min = self.config_limit_gain * Delta_q_min[self.indices]
        G = np.vstack([self.projection_matrix, -self.projection_matrix])
        h = np.hstack([p_max, -p_min])
        return G, h



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Upkie wheeled biped bending its knees."""

import numpy as np
import pinocchio as pin
import qpsolvers
import yourdfpy
from loop_rate_limiters import RateLimiter

import pink
from pink import solve_ik
from pink.tasks import FrameTask, PostureTask
from pink.utils import custom_configuration_vector

try:
    from robot_descriptions import upkie_description
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try `pip install robot_descriptions`"
    )


if __name__ == "__main__":
    robot = load_robot_description(
        "upkie_description", root_joint=pin.JointModelFreeFlyer()
    )

    tasks = {
        "base": FrameTask(
            "base",
            position_cost=1.0,  # [cost] / [m]
            orientation_cost=1.0,  # [cost] / [rad]
        ),
        "left_contact": FrameTask(
            "left_contact",
            position_cost=[0.1, 0.0, 0.1],  # [cost] / [m]
            orientation_cost=0.0,  # [cost] / [rad]
        ),
        "right_contact": FrameTask(
            "right_contact",
            position_cost=[0.1, 0.0, 0.1],  # [cost] / [m]
            orientation_cost=0.0,  # [cost] / [rad]
        ),
        "posture": PostureTask(
            cost=1e-3,  # [cost] / [rad]
        ),
    }

    configuration = pink.Configuration(robot.model, robot.data, robot.q0)
    for body, task in tasks.items():
        if type(task) is FrameTask:
            task.set_target_from_configuration(configuration)

    tasks["posture"].set_target(
        custom_configuration_vector(robot, left_knee=0.2, right_knee=-0.2)
    )

    left_contact_target = configuration.get_transform_frame_to_world(
        "left_contact"
    )
    right_contact_target = configuration.get_transform_frame_to_world(
        "right_contact"
    )

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "quadprog" in qpsolvers.available_solvers:
        solver = "quadprog"

    animation_time = 0.0  # [s]
    visualizer_fps = 100  # [Hz]
    rate = RateLimiter(frequency=visualizer_fps)

    def callback(scene, dz=0.05):
        """Callback function for the visualizer."""
        global animation_time, configuration
        dt = rate.period

        # Update task targets
        t = animation_time
        left_contact_target.translation[2] += 0.1 * np.sin(t) * dt
        right_contact_target.translation[2] += 0.1 * np.sin(t) * dt
        tasks["left_contact"].set_target(left_contact_target)
        tasks["right_contact"].set_target(right_contact_target)

        # Compute velocity and integrate it into next configuration
        velocity = solve_ik(configuration, tasks.values(), dt, solver=solver)
        configuration.integrate_inplace(velocity, dt)

        # Display resulting configuration
        actuated_joints = configuration.q[7:]
        viz.update_cfg(actuated_joints)

        # Regulate visualizer FPS
        animation_time += dt
        rate.sleep()

    viz = yourdfpy.URDF.load(upkie_description.URDF_PATH)
    viz.show(callback=callback)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2024 Inria

"""Relative frame task implementation."""

from typing import Optional, Sequence, Union

import numpy as np
import pinocchio as pin

from ..configuration import Configuration
from .exceptions import TargetNotSet, TaskDefinitionError
from .task import Task


class RelativeFrameTask(Task):
    r"""Regulate the pose of a robot frame relative to another robot frame.

    Attributes:
        frame: Frame name, typically the name of a link or joint from the robot
            description.
        root: Name of the robot frame the task is relative to.
        transform_target_to_root: Target pose of the frame.
    """

    cost: np.ndarray
    frame: str
    root: str
    transform_target_to_root: Optional[pin.SE3]

    def __init__(
        self,
        frame: str,
        root: str,
        position_cost: Union[float, Sequence[float]],
        orientation_cost: Union[float, Sequence[float]],
        lm_damping: float = 0.0,
        gain: float = 1.0,
    ) -> None:
        r"""Define a relative frame task.

        Args:
            frame: Frame name, typically the name of a link or joint from the
                robot description.
            root: Name of the robot frame the task is relative to.
            position_cost: Contribution of position errors to the normalized
                cost, in :math:`[\mathrm{cost}] / [\mathrm{m}]`. If this is a
                vector, the cost is anisotropic and each coordinate corresponds
                to an axis of the frame.
            orientation_cost: Contribution of orientation errors to the
                normalized cost, in :math:`[\mathrm{cost}] / [\mathrm{rad}]`.
                If this is a vector, the cost is anisotropic and each
                coordinate corresponds to an axis of the frame.
            lm_damping: Levenberg-Marquardt damping (see class attributes). The
                default value is conservatively low.
            gain: Task gain :math:`\alpha \in [0, 1]` for additional low-pass
                filtering. Defaults to 1.0 (no filtering) for dead-beat
                control.
        """
        super().__init__(
            cost=np.ones(6),  # updated below
            gain=gain,
            lm_damping=lm_damping,
        )
        self.frame = frame
        self.root = root
        self.lm_damping = lm_damping
        self.transform_target_to_root = None
        #
        self.set_position_cost(position_cost)
        self.set_orientation_cost(orientation_cost)

    def set_position_cost(
        self, position_cost: Union[float, Sequence[float], np.ndarray]
    ) -> None:
        r"""Set a new cost for all 3D position coordinates.

        Args:
            position_cost: Contribution of position errors to the normalized
                cost, in :math:`[\mathrm{cost}] / [\mathrm{m}]`. If this is a
                vector, the cost is anisotropic and each coordinate corresponds
                to an axis of the frame.
        """
        if isinstance(position_cost, float):
            assert position_cost >= 0.0
        else:  # Should be a ndarray or seq
            if not isinstance(position_cost, np.ndarray):  # Must be seq
                try:
                    position_cost = np.array(position_cost)
                except Exception:  # Not a proper float sequence
                    raise TaskDefinitionError(
                        "Position task cost should be a float or a "
                        "seq of float or ndarray of size 1 or 3,"
                        f"currently cost={self.cost}"
                    )
            assert np.all(position_cost >= 0.0)

        self.cost[0:3] = position_cost

    def set_orientation_cost(
        self, orientation_cost: Union[float, Sequence[float], np.ndarray]
    ) -> None:
        r"""Set a new cost for all 3D orientation coordinates.

        Args:
            orientation_cost: Contribution of orientation errors to the
                normalized cost, in :math:`[\mathrm{cost}] / [\mathrm{rad}]`.
                If this is a vector, the cost is anisotropic and each
                coordinate corresponds to an axis of the frame.
        """
        if isinstance(orientation_cost, float):
            assert orientation_cost >= 0.0
        else:  # Should be a ndarray or seq
            if not isinstance(orientation_cost, np.ndarray):  # Must be seq
                try:
                    orientation_cost = np.array(orientation_cost)
                except Exception:  # Not a proper float sequence
                    raise TaskDefinitionError(
                        "Orientation task cost should be a float or a "
                        "seq of float or ndarray of size 1 or 3,"
                        f"currently cost={self.cost}"
                    )
            assert np.all(orientation_cost >= 0.0)

        self.cost[3:] = orientation_cost

    def set_target(
        self,
        transform_target_to_root: pin.SE3,
    ) -> None:
        """Set task target pose in the root frame.

        Args:
            transform_target_to_root: Affine transform matrix (4×4) from the
                task target frame to the root frame.
        """
        self.transform_target_to_root = transform_target_to_root.copy()

    def set_target_from_configuration(
        self, configuration: Configuration
    ) -> None:
        """Set task target pose from a robot configuration.

        Args:
            configuration: Robot configuration.
        """
        self.set_target(configuration.get_transform(self.frame, self.root))

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the error twist.

        The error is a twist :math:`e(q) \in se(3)` expressed in the local
        frame (i.e., it is a *body* twist). We map it to :math:`\mathbb{R}^6`
        following Pinocchio's convention where linear coordinates are followed
        by angular coordinates. The error vector is:

        .. math::

            e(q) := \log(T_{tf}) = \log(T_{rt}^{-1} T_{rf})

        where :math:`b` denotes our frame, :math:`t` the target frame and
        :math:`0` the inertial frame. See also
        :func:`pink.tasks.task.Task.compute_error` for more context.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Frame task error :math:`e(q)`.
        """
        if self.transform_target_to_root is None:
            raise TargetNotSet(
                f"target pose of frame '{self.frame}' "
                f"in frame '{self.root}' is undefined"
            )
        transform_frame_to_root = configuration.get_transform(
            self.frame, self.root
        )
        transform_frame_to_target = self.transform_target_to_root.actInv(
            transform_frame_to_root
        )
        error_in_frame: np.ndarray = pin.log(transform_frame_to_target).vector
        return error_in_frame

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the frame task Jacobian.

        The task Jacobian :math:`J(q) \in \mathbb{R}^{6 \times n_v}` is the
        derivative of the task error :math:`e(q) \in \mathbb{R}^6` with respect
        to the configuration :math:`q`. The formula for the frame task is:

        .. math::

            J(q) = \text{Jlog}(T_{tf}) (
                {}_f J_{0f}(q) - \text{Ad}_{T_{fr}}{}_r J_{0r}(q)
            )

        The proof is as follow if we denote :math:`JT(q)` the Jacobian in the
        local frame :math:`T_{tf}(q) = T_{rt}^{-1} T_{0r}^{-1}(q) T_{0f}(q)`
        we have

        .. math::
            \begin{align}
            J(q) = \text{Jlog}(T_{tf}) JT
            [Jv]^{up} & = T_{tf}^{-1} \dot{T}_{tf} \\
                & = T_{ft}T_{tr}T_{r0}\dot{T}_{0f}
                - T_{ft}T_{tr}T_{0r}^{-1}\dot{T}_{0r}T_{0r}^{-1}T_{0f}
                \\
                & = T_{0f}^{-1}\dot{T}_{0f}
                - T_{fr}(T_{0r}^{-1}\dot{T}_{0r})T_{fr}^{-1} \\
                & = [{}_f J_{0f} v - \text{Ad}_{T_{fr}}{}_r J_{0r} v]^{up} \\
            J(q) & = {}_f J_{0f} - \text{Ad}_{T_{fr}}{}_r J_{0r}
            \end{align}

        The formula implemented here is more general than the one detailed in
        [FrameTaskJacobian]_. See also
        :func:`pink.tasks.task.Task.compute_jacobian` for more context on task
        Jacobians.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Jacobian matrix :math:`J`, expressed locally in the frame.
        """
        if self.transform_target_to_root is None:
            raise TargetNotSet(
                f"target pose of frame '{self.frame}' "
                f"in frame '{self.root}' is undefined"
            )
        transform_frame_to_root = configuration.get_transform(
            self.frame, self.root
        )
        transform_frame_to_target = self.transform_target_to_root.actInv(
            transform_frame_to_root
        )
        action_root_to_frame = transform_frame_to_root.actionInverse
        jacobian_frame_in_frame = configuration.get_frame_jacobian(self.frame)
        jacobian_root_in_root = configuration.get_frame_jacobian(self.root)
        J = pin.Jlog6(transform_frame_to_target) @ (
            jacobian_frame_in_frame
            - action_root_to_frame @ jacobian_root_in_root
        )
        return J

    def __repr__(self):
        """Human-readable representation of the task."""
        orientation_cost = (
            self.cost if isinstance(self.cost, float) else self.cost[3:6]
        )
        position_cost = (
            self.cost if isinstance(self.cost, float) else self.cost[0:3]
        )
        return (
            "RelativeFrameTask("
            f"frame={self.frame}, "
            f"root={self.root}, "
            f"gain={self.gain}, "
            f"orientation_cost={orientation_cost}, "
            f"position_cost={position_cost}, "
            f"transform_target_to_root={self.transform_target_to_root})"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0

"""Move an end-effector to a prescribed target."""

import numpy as np
import pinocchio
from robot_descriptions.loaders.pinocchio import load_robot_description

import pink

robot = load_robot_description("ur10_description")

# Frame details
FRAME_NAME = "ee_frame"
joint_name = robot.model.names[-1]
parent_joint = robot.model.getJointId(joint_name)
parent_frame = robot.model.getFrameId(joint_name)
placement = pinocchio.SE3.Identity()

ee_frame = robot.model.addFrame(
    pinocchio.Frame(
        FRAME_NAME,
        parent_joint,
        parent_frame,
        placement,
        pinocchio.FrameType.OP_FRAME,
    )
)
robot.data = pinocchio.Data(robot.model)
low = robot.model.lowerPositionLimit
high = robot.model.upperPositionLimit
robot.q0 = pinocchio.neutral(robot.model)

# Task details
np.random.seed(0)
q_final = np.array(
    [
        np.random.uniform(low=low[i], high=high[i], size=(1,))[0]
        for i in range(6)
    ]
)
pinocchio.forwardKinematics(robot.model, robot.data, q_final)

target_pose = robot.data.oMi[parent_joint]
ee_task = pink.tasks.FrameTask(FRAME_NAME, [1.0, 1.0, 1.0], [1.0, 1.0, 1.0])
ee_task.set_target(target_pose)
tasks = [ee_task]

# Inverse kinematics parameters
dt = 1e-2
damping = 1e-8
n_iter = 10000
solver = "quadprog"

configuration = pink.Configuration(robot.model, robot.data, robot.q0)

for i in range(n_iter):
    dv = pink.solve_ik(
        configuration,
        [ee_task],
        dt=dt,
        damping=damping,
        solver=solver,
    )
    q_out = pinocchio.integrate(robot.model, configuration.q, dv * dt)
    configuration = pink.Configuration(robot.model, robot.data, q_out)
    pinocchio.updateFramePlacements(robot.model, robot.data)
    err = ee_task.compute_error(configuration)
    print(i, err)
    if np.linalg.norm(ee_task.compute_error(configuration)) < 1e-8:
        break



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Ivan Domrachev, Simeon Nedelchev

"""All barriers derive from the :class:`Barrier` base class.

The formalism used in this implementation is written down in
https://simeon-ned.com/blog/2024/cbf/
"""

import abc
from typing import Callable, Optional, Tuple, Union

import numpy as np

from ..configuration import Configuration


class Barrier(abc.ABC):
    r"""Abstract base class for barrier.

    A barrier is a function :math:`h(q)` that
    satisfies the following condition:

    .. math::

        \frac{\partial h_j}{\partial q}
        \dot{q} +\alpha_j(h_j(q))
        \geq 0, \quad \forall j

    where :math:`\frac{\partial h_j}{\partial q}`
    are the Jacobians of the constraint functions, :math:`\dot{q}`
    is the joint velocity vector, and :math:`\alpha_j` are extended
    `class kappa <https://en.wikipedia.org/wiki/Class_kappa_function>`__
    functions.

    On top of that, following `this article
    <https://arxiv.org/pdf/2404.12329>`__ barriers utilize safe displacement
    term is added to the cost of the optimization problem:

    .. math::

        \frac{r}{2\|J_h\|^{2}}\|dq-dq_{\text{safe}}(q)\|^{2},

    where :math:`J_h` is the Jacobian of the barrier function, dq is the
    joint displacement vector, and :math:`dq_{\text{safe}}(q)` is the safe
    displacement vector.

    Attributes:
        dim: Dimension of the barrier.
        gain: linear barrier gain.
        gain_function: function, that defines stabilization term as nonlinear
            function of barrier. Defaults to the (linear) identity function.
        safe_displacement: Safe backup displacement.
        safe_displacement_gain: positive gain for safe backup displacement.
    """

    gain: np.ndarray
    gain_function: Callable[[float], float]
    safe_displacement: np.ndarray
    safe_displacement_gain: float

    def __init__(
        self,
        dim: int,
        gain: Union[float, np.ndarray] = 1.0,
        gain_function: Optional[Callable[[float], float]] = None,
        safe_displacement_gain: float = 0.0,
    ):
        """Initialize the barrier.

        Args:
            dim: Dimension of the barrier.
            gain: barrier gain. Defaults to 1.0.
            gain_function: gain function.
                Defaults to the identity function.
            safe_displacement_gain: gain for the safe backup displacement
                cost term. Defaults to 3.0.
        """
        self.dim = dim
        self.gain = (
            gain if isinstance(gain, np.ndarray) else np.ones(dim) * gain
        )

        self.gain_function = (
            gain_function if gain_function is not None else lambda x: x
        )
        self.safe_displacement = np.zeros(self.dim)
        self.safe_displacement_gain = safe_displacement_gain

        # Cached values to avoid recomputation
        self.__q_cache: Optional[np.ndarray] = None
        self.__jac_cache: Optional[np.ndarray] = None
        self.__barrier_cache: Optional[np.ndarray] = None

    @abc.abstractmethod
    def compute_barrier(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the value of the barrier function.

        The barrier function :math:`h(q)`
        is a vector-valued function that represents the safety constraints.
        It should be designed such that the set
        :math:`\{q : h(q) \geq 0\}`
        represents the safe region of the configuration space.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Value of the barrier function :math:`h(q)`.
        """

    @abc.abstractmethod
    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the Jacobian matrix of the barrier function.

        The Jacobian matrix
        :math:`\frac{\partial h}{\partial q}(q)`
        of the barrier function with respect to the configuration variables is
        required for the computation of the barrier condition.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Jacobian matrix
            :math:`\frac{\partial h}{\partial q}(q)`.
        """

    def compute_safe_displacement(
        self, configuration: Configuration
    ) -> np.ndarray:
        r"""Compute the safe backup displacement.

        The safe backup control displacement :math:`dq_{safe}(q)`
        is a joint displacement vector that can guarantee that system
        would stay in safety set.

        By default, it is set to zero, since it could not violate safety set.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Safe backup joint velocities
                :math:`\dot{q}_{safe}(q)`.
        """
        return np.zeros(configuration.model.nv)

    def compute_qp_objective(
        self,
        configuration: Configuration,
    ) -> Tuple[np.ndarray, np.ndarray]:
        r"""Compute the quadratic objective function for the barrier-based QP.

        The quadratic objective function includes a regularization term
        based on the safe backup policy:

        .. math::

            \gamma(q)\left\| \dot{q}-
            \dot{q}_{safe}(q)\right\|^{2}

        where :math:`\gamma(q)` is a configuration-dependent weight and
        :math:`\dot{q}_{safe}(q)` is the safe backup policy.

        Note:
            If `safe_displacement_gain` is set to zero, the regularization
            term is not included. Jacobian and barrier values are cached
            to avoid recomputation.

        Args:
            configuration: Robot configuration :math:`q`.
            dt: Time step for discrete-time implementation. Defaults to 1e-3.

        Returns:
            Tuple containing the quadratic objective matrix (H) and linear
                objective vector (c).
        """
        if (
            self.__q_cache is None
            or self.__jac_cache is None
            or self.__barrier_cache is None
        ) or not np.allclose(self.__q_cache, configuration.q):
            self.__q_cache = configuration.q
            self.__jac_cache = self.compute_jacobian(configuration)
            self.__barrier_cache = self.compute_barrier(configuration)

        H = np.zeros((configuration.model.nv, configuration.model.nv))
        c = np.zeros(configuration.model.nv)

        if self.safe_displacement_gain > 1e-6:
            self.safe_displacement = self.compute_safe_displacement(
                configuration
            )
            jac_squared_norm = np.linalg.norm(self.__jac_cache) ** 2
            gain_over_jacobian = self.safe_displacement_gain / jac_squared_norm

            H += gain_over_jacobian * np.eye(configuration.model.nv)
            c += -gain_over_jacobian * self.safe_displacement

        return (H, c)

    def compute_qp_inequalities(
        self,
        configuration: Configuration,
        dt: float = 1e-3,
    ) -> Tuple[np.ndarray, np.ndarray]:
        r"""Compute the linear inequality constraints for the barrier-based QP.

        The linear inequality constraints enforce the barrier conditions:

        .. math::

            \frac{\partial h_j}
            {\partial q} \dot{q} +
            \alpha_j(h_j(q)) \geq 0, \quad \forall j

        where :math:`\frac{\partial h_j}{\partial q}`
        are the Jacobians of the constraint functions,
        :math:`\dot{q}` is the joint velocity vector,
        and :math:`\alpha_j` are extended class K functions.

        Note:
            Jacobian and barrier values are cached to avoid recomputation.

        Args:
            configuration: Robot configuration :math:`q`.
            dt: Time step for discrete-time implementation. Defaults to 1e-3.

        Returns:
            Tuple containing the inequality constraint matrix (G)
                and vector (h).
        """
        if (
            self.__q_cache is None
            or self.__jac_cache is None
            or self.__barrier_cache is None
            or not np.allclose(self.__q_cache, configuration.q)
        ):
            self.__q_cache = configuration.q
            self.__jac_cache = self.compute_jacobian(configuration)
            self.__barrier_cache = self.compute_barrier(configuration)

        G = -self.__jac_cache / dt
        h = np.array(
            [
                self.gain[i] * self.gain_function(self.__barrier_cache[i])
                for i in range(self.dim)
            ]
        )

        return (G, h)

    def __repr__(self) -> str:
        """Return a string representation of the barrier.

        Returns:
            str: String representation.
        """
        return (
            f"Barrier("
            f"gain={self.gain}, "
            f"safe_displacement={'no' if np.allclose(self.safe_displacement_gain, 0) else self.safe_displacement}, "  # noqa: E501
            f"safe_displacement_gain={'no' if np.allclose(self.safe_displacement_gain, 0) else self.safe_displacement_gain})"  # noqa: E501
            f"dim={self.dim})"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Joint coupling task implementation."""

from typing import Sequence

import numpy as np
import pinocchio as pin

from ..configuration import Configuration
from .linear_holonomic_task import LinearHolonomicTask


class JointCouplingTask(LinearHolonomicTask):
    r"""Coupling between two revolute joints.

    Note:
        This task only considers a 1-Dimensional task.
    """

    joint_names: Sequence[str]

    def __init__(
        self,
        joint_names: Sequence[str],
        ratios: Sequence[float],
        cost: float,
        configuration: Configuration,
        lm_damping: float = 0.0,
        gain: float = 1.0,
    ) -> None:
        r"""Compute Jacobian matrix of a linear holonomic constraint.

        Args:
            joint_names: a list of joint names consisting of a linear
                holonomic constraint.
            ratios: a list of ratios that each joint takes in a linear
                holonomic constraint.
            cost: value used to cast joint angle differences to a cost, in
                :math:`[\mathrm{cost}] / [\mathrm{rad}]`.
            configuration: Configuration used to retrieve joint indices from
                joint names.
            lm_damping: Unitless scale of the Levenberg-Marquardt (only when
                the error is large) regularization term, which helps when
                targets are unfeasible. Increase this value if the task is too
                jerky under unfeasible targets, but beware that too large a
                damping can slow down the task.
            gain: Task gain :math:`\alpha \in [0, 1]` for additional low-pass
                filtering. Defaults to 1.0 (no filtering) for dead-beat
                control.
        """
        assert len(joint_names) == len(ratios)

        A = np.zeros((1, configuration.model.nv))
        for joint, ratio in zip(joint_names, ratios):
            joint_obj = configuration.model.joints[
                configuration.model.getJointId(joint)
            ]
            strt = joint_obj.idx_v
            end = strt + joint_obj.nv
            A[:, strt:end] = ratio

        super().__init__(
            A,
            np.zeros(1),
            pin.neutral(configuration.model),
            cost=cost,
            gain=gain,
            lm_damping=lm_damping,
        )
        self.joint_names = joint_names
        self.ratios = ratios

    def __repr__(self):
        """Human-readable representation of the task."""
        return (
            "JointCouplingTask("
            f"joint_names={self.joint_names}, "
            f"ratios={self.ratios}, "
            f"cost={self.cost}, "
            f"gain={self.gain}, "
            f"lm_damping={self.lm_damping})"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""Test fixture for the low-acceleration task."""

import unittest

import numpy as np
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.tasks import LowAccelerationTask


class TestLowAccelerationTask(unittest.TestCase):
    """Test consistency of the low-acceleration task."""

    def setUp(self):
        """Prepare test fixture."""
        robot = load_robot_description("ur3_description", root_joint=None)
        self.configuration = Configuration(robot.model, robot.data, robot.q0)

    def test_task_repr(self):
        """String representation reports the task gain, costs and target."""
        task = LowAccelerationTask(cost=1.0)
        self.assertTrue("cost=" in repr(task))
        self.assertFalse("gain=" in repr(task))
        self.assertFalse("lm_damping=" in repr(task))

    def test_qp_objective(self):
        task = LowAccelerationTask(cost=1.0)
        nv = self.configuration.model.nv
        v_prev = np.array([1.0, 2.0, 3.0, 4.0, -3.0, -2.0])
        dt = 1.234e-2  # [s]
        task.set_last_integration(v_prev, dt)
        H, c = task.compute_qp_objective(self.configuration)
        self.assertLess(np.linalg.norm(H - np.eye(nv)), 1e-10)
        self.assertLess(np.linalg.norm(c + v_prev * dt), 1e-10)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""All kinematic tasks derive from the :class:`Task` base class.

The formalism used in this implementation is written down in `this note on
task-based inverse kinematics
<https://scaron.info/robotics/inverse-kinematics.html>`_. As of February 2022,
it hasn't been updated with the proper dimensional analysis, but the core
concepts and notations are there.
"""

import abc
from typing import Optional, Sequence, Tuple, Union

import numpy as np

from ..configuration import Configuration


class Task(abc.ABC):
    r"""Abstract base class for kinematic tasks.

    Attributes:
        cost: cost vector with the same dimension as the error of the task. Its
            units depends on the error as well.
        gain: Task gain :math:`\alpha \in [0, 1]` for additional low-pass
            filtering. Defaults to 1.0 (no filtering) for dead-beat control.
        lm_damping: Unitless scale of the Levenberg-Marquardt (only when
            the error is large) regularization term, which helps when
            targets are unfeasible. Increase this value if the task is too
            jerky under unfeasible targets, but beware that a larger
            damping slows down the task.
    """

    cost: Optional[Union[float, Sequence[float], np.ndarray]]
    gain: float
    lm_damping: float

    def __init__(
        self,
        cost: Optional[Union[float, Sequence[float], np.ndarray]] = None,
        gain: float = 1.0,
        lm_damping: float = 0.0,
    ):
        r"""Create task.

        Args:
            cost: cost vector with the same dimension as the error of the task.
                Its units depends on the error as well.
            gain: Task gain :math:`\alpha \in [0, 1]` for additional low-pass
                filtering. Defaults to 1.0 (no filtering) for dead-beat
                control.
            lm_damping: Unitless scale of the Levenberg-Marquardt (only when
                the error is large) regularization term, which helps when
                targets are unfeasible. Increase this value if the task is too
                jerky under unfeasible targets, but beware that too large a
                damping can slow down the task.
        """
        self.cost = cost
        self.gain = gain
        self.lm_damping = lm_damping

    @abc.abstractmethod
    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the task error function.

        The error function :math:`e(q) \in \mathbb{R}^{k}` is the quantity that
        the task aims to drive to zero (:math:`k` is the dimension of the
        task). It appears in the first-order task dynamics:

        .. math::

            J(q) \Delta q = -\alpha e(q)

        The Jacobian matrix :math:`J(q) \in \mathbb{R}^{k \times n_v}`, with
        :math:`n_v` the dimension of the robot's tangent space, is the
        derivative of the task error :math:`e(q)` with respect to the
        configuration :math:`q \in \mathbb{R}^{n_q}`. This Jacobian is
        implemented in :func:`Task.compute_jacobian`. Finally, the
        configuration displacement :math:`\Delta q` is the output of inverse
        kinematics.

        In the first-order task dynamics, the error :math:`e(q)` is multiplied
        by the task gain :math:`\alpha \in [0, 1]`. This gain can be one for
        dead-beat control (*i.e.* converge as fast as possible, but might be
        unstable as it neglects our first-order approximation), but it can also
        be lower a slower task (similar to low-pass filtering).

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Task error vector :math:`e(q)`.
        """

    @abc.abstractmethod
    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the task Jacobian at a given configuration.

        The task Jacobian :math:`J(q) \in \mathbb{R}^{k \times n_v}` is the
        first-order derivative of the error :math:`e(q) \in \mathbb{R}^{k}`
        that defines the task, with :math:`k` the dimension of the task and
        :math:`n_v` the dimension of the robot's tangent space.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Task Jacobian :math:`J(q)`.
        """

    def compute_qp_objective(
        self, configuration: Configuration
    ) -> Tuple[np.ndarray, np.ndarray]:
        r"""Compute the matrix-vector pair :math:`(H, c)` of the QP objective.

        This pair is such that the contribution of the task to the QP objective
        of the IK is:

        .. math::

            \| J \Delta q + \alpha e \|_{W}^2 = \frac{1}{2} \Delta q^T H
            \Delta q + c^T q

        The weight matrix :math:`W \in \mathbb{R}^{k \times k}` weighs and
        normalizes task coordinates to the same unit. The unit of the overall
        contribution is [cost]^2. The configuration displacement :math:`\Delta
        q` is the output of inverse kinematics (we divide it by :math:`\Delta
        t` to get a commanded velocity).

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Pair :math:`(H(q), c(q))` of Hessian matrix and linear vector of
            the QP objective.

        See Also:
            Levenberg-Marquardt damping is described in [Sugihara2011]_. The
            dimensional analysis in this class is our own.
        """
        jacobian = self.compute_jacobian(configuration)
        minus_gain_error = -self.gain * self.compute_error(configuration)

        weight = (
            np.eye(jacobian.shape[0])
            if self.cost is None
            else np.diag(
                [self.cost] * jacobian.shape[0]
                if isinstance(self.cost, float)
                else self.cost
            )
        )

        weighted_jacobian = weight @ jacobian  # [cost]
        weighted_error = weight @ minus_gain_error  # [cost]
        mu = self.lm_damping * weighted_error @ weighted_error  # [cost]^2
        eye_tg = configuration.tangent.eye
        # Our Levenberg-Marquardt damping `mu * eye_tg` is isotropic in the
        # robot's tangent space. If it helps we can add a tangent-space scaling
        # to damp the floating base differently from joint angular velocities.
        H = weighted_jacobian.T @ weighted_jacobian + mu * eye_tg
        c = -weighted_error.T @ weighted_jacobian
        return (H, c)

    def __repr__(self):
        """Human-readable representation of the task."""
        return (
            f"Task("
            f"cost={self.cost}, "
            f"gain={self.gain}, "
            f"lm_damping={self.lm_damping})"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria

"""Move a Stretch RE1 with a fingertip target in the mobile-base frame."""

import meshcat_shapes
import numpy as np
import pinocchio as pin
import qpsolvers
from loop_rate_limiters import RateLimiter

import pink
from pink import solve_ik
from pink.tasks import FrameTask, RelativeFrameTask
from pink.visualization import start_meshcat_visualizer

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try ``pip install robot_descriptions``"
    )

# Trajectory parameters to play with ;)
CIRCLE_RADIUS = 0.5  # [m]
FINGERTIP_HEIGHT = 0.7  # [m]

if __name__ == "__main__":
    robot = load_robot_description(
        "stretch_description", root_joint=pin.JointModelPlanar()
    )
    # __import__("IPython").embed()

    # Initialize visualizer
    viz = start_meshcat_visualizer(robot)
    viewer = viz.viewer
    meshcat_shapes.frame(viewer["base_target_frame"], opacity=0.5)
    meshcat_shapes.frame(viewer["fingertip_target_frame"], opacity=1.0)

    # Define tasks
    base_task = FrameTask(
        "base_link",
        position_cost=0.1,  # [cost] / [m]
        orientation_cost=1.0,  # [cost] / [rad]
    )
    fingertip_task = RelativeFrameTask(
        "link_gripper_fingertip_right",
        "base_link",
        position_cost=1.0,
        orientation_cost=1e-4,
    )
    tasks = [base_task, fingertip_task]

    # Initialize tasks from the initial configuration
    configuration = pink.Configuration(robot.model, robot.data, robot.q0)
    base_task.set_target_from_configuration(configuration)
    transform_fingertip_target_to_base = pin.SE3(
        rotation=np.eye(3),
        translation=np.array([0.1, 0.0, FINGERTIP_HEIGHT]),
    )
    transform_fingertip_to_world = configuration.get_transform_frame_to_world(
        fingertip_task.frame
    )
    center_translation = transform_fingertip_to_world.translation[:2]
    fingertip_task.set_target(transform_fingertip_target_to_base)
    viz.display(configuration.q)

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "quadprog" in qpsolvers.available_solvers:
        solver = "quadprog"

    rate = RateLimiter(frequency=100.0)
    dt = rate.period
    t = 0.0  # [s]
    while True:
        # Update base task target
        T = base_task.transform_target_to_world
        u = np.array([np.cos(t), np.sin(t)])
        T.translation[:2] = center_translation + CIRCLE_RADIUS * u
        T.rotation = pin.utils.rpyToMatrix(0.0, 0.0, 0.5 * np.pi * t)

        # Update fingertip target
        fingertip_task.transform_target_to_root.translation[2] = (
            FINGERTIP_HEIGHT + 0.1 * u[1]
        )

        # Update visualizer frames
        viewer["base_target_frame"].set_transform(T.np)
        transform_fingertip_target_to_world = (
            configuration.get_transform_frame_to_world("base_link")
            * fingertip_task.transform_target_to_root
        )
        viewer["fingertip_target_frame"].set_transform(
            transform_fingertip_target_to_world.np
        )
        viewer["base_frame"].set_transform(
            configuration.get_transform_frame_to_world(base_task.frame).np
        )
        viewer["fingertip_frame"].set_transform(
            configuration.get_transform_frame_to_world(fingertip_task.frame).np
        )

        # Compute velocity and integrate it into next configuration
        velocity = solve_ik(configuration, tasks, dt, solver=solver)
        configuration.integrate_inplace(velocity, dt)

        # Visualize result at fixed FPS
        viz.display(configuration.q)
        rate.sleep()
        # input("next?")
        t += dt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Tests for position barrier limit."""

import unittest

import numpy as np
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.barriers import PositionBarrier, NoPositionLimitProvided


class TestPositionBarrier(unittest.TestCase):
    """Testing position barrier limit."""

    def setUp(self):
        """Set test fixture up."""
        robot = load_robot_description("ur3_description")
        model = robot.model
        self.data = robot.data
        self.model = model
        self.robot = robot
        self.configuration = Configuration(
            self.model, self.data, np.zeros(self.model.nq)
        )

    def test_no_position_limit(self):
        """Raise an error when the robot body is not found."""
        with self.assertRaises(NoPositionLimitProvided):
            PositionBarrier("foo")

    def test_dimension(self):
        """Check dimensions of configuration limit projection."""
        self.assertEqual(PositionBarrier("ee_link", p_min=np.zeros(3)).dim, 3)
        self.assertEqual(PositionBarrier("ee_link", p_max=np.zeros(3)).dim, 3)
        self.assertEqual(
            PositionBarrier(
                "ee_link", p_min=np.zeros(3), p_max=np.zeros(3)
            ).dim,
            6,
        )

    def test_gains(self):
        """Check gains of configuration limit projection."""
        # One limit, scalar gain
        self.assertEqual(
            PositionBarrier("ee_link", p_min=np.zeros(3), gain=1).gain.shape,
            (3,),
        )
        # One limit, vector gain
        self.assertEqual(
            PositionBarrier(
                "ee_link", p_max=np.zeros(3), gain=np.array([1, 2, 3])
            ).gain.shape,
            (3,),
        )
        # Two limits, scalar gain
        self.assertEqual(
            PositionBarrier(
                "ee_link", p_min=np.zeros(3), p_max=np.zeros(3), gain=1
            ).gain.shape,
            (6,),
        )
        # Two limits, vector gain
        self.assertEqual(
            PositionBarrier(
                "ee_link",
                p_min=np.zeros(3),
                p_max=np.zeros(3),
                gain=np.array([1, 2, 3]),
            ).gain.shape,
            (6,),
        )

    def test_positive_when_in_safety_zone(self):
        """Check that the barrier is positive when in the safety zone."""

        barrier = PositionBarrier("ee_link", p_min=np.zeros(3))
        h = barrier.compute_barrier(self.configuration)
        self.assertTrue(np.all(h > 0))

    def test_negative_when_out_of_safety_zone(self):
        """Test that the barrier is negative when out of the safety zone."""

        for violated_idx in range(3):
            p_min = np.zeros(3)
            p_min[violated_idx] = 1.0
            barrier = PositionBarrier("ee_link", p_min=p_min)
            h = barrier.compute_barrier(self.configuration)
            self.assertTrue(np.any(h < 0))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""All kinematic limits derive from the :class:`Limit` base class."""

import abc
from typing import Optional, Tuple

import numpy as np


class Limit(abc.ABC):
    r"""Abstract base class for kinematic limits."""

    @abc.abstractmethod
    def compute_qp_inequalities(
        self,
        q: np.ndarray,
        dt: float,
    ) -> Optional[Tuple[np.ndarray, np.ndarray]]:
        r"""Compute limit as linearized QP inequalities.

        Those limits are returned as:

        .. math::

            G(q) \Delta q \leq h(q)

        where :math:`q \in {\cal C}` is the robot's configuration and
        :math:`\Delta q \in T_q({\cal C})` is the displacement in the tangent
        space at :math:`q`.

        Args:
            q: Robot configuration.
            dt: Integration timestep in [s].

        Returns:
            Pair :math:`(G, h)` representing the inequality constraint as
            :math:`G \Delta q \leq h`, or ``None`` if there is no limit.
        """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""A pair of UR3 arms on a mobile body tracking moving targets."""

from typing import Tuple

import hppfcl as fcl
import meshcat_shapes
import numpy as np
import pinocchio as pin
import qpsolvers
from loop_rate_limiters import RateLimiter

import pink
from pink import solve_ik
from pink.tasks import FrameTask, PostureTask
from pink.utils import custom_configuration_vector
from pink.visualization import start_meshcat_visualizer

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try ``pip install robot_descriptions``"
    ) from exc


def create_floating_base(
    name: str = "base",
    density: float = 1e-3,
    width: float = 0.8,
    height: float = 0.4,
    depth: float = 0.2,
) -> Tuple[pin.Model, pin.GeometryModel]:
    """Creating a floating base body to hold both arms.

    Args:
        name: Name of the floating base body.
        density: Mass density in kg / m³.
        width: Body width in meters.
        height: Body height in meters.
        depth: Body depth in meters.

    Returns:
        Joint and visual models.
    """
    model = pin.Model()
    mass = density * width * height * depth
    body_inertia = pin.Inertia.FromBox(mass, width, height, depth)
    parent = 0  # universe
    joint_id = model.addJoint(
        parent, pin.JointModelFreeFlyer(), pin.SE3.Identity(), name
    )
    model.appendBodyToJoint(joint_id, body_inertia, pin.SE3.Identity())
    model.addJointFrame(joint_id, -1)
    model.addBodyFrame(f"{name}_body", joint_id, pin.SE3.Identity(), -1)

    visual_model = pin.GeometryModel()
    box = fcl.Box(width, height, depth)
    box_placement = pin.SE3.Identity()
    box_placement.translation = np.array(
        [-0.0 * width, -0.0 * height, -0.5 * depth]
    )
    geom_obj = pin.GeometryObject("box", joint_id, box, box_placement)
    geom_obj.meshColor = np.array([0.9, 0.8, 0.6, 1.0])
    visual_model.addGeometryObject(geom_obj)
    return model, visual_model


def prefix_frames(
    model: pin.Model, visual_model: pin.GeometryModel, prefix: str
) -> None:
    for frame in model.frames:
        frame.name = f"{prefix}_{frame.name}"
    for i, name in enumerate(model.names):
        model.names[i] = f"{prefix}_{name}"
    for geom in visual_model.geometryObjects:
        geom.name = f"{prefix}_{geom.name}"


if __name__ == "__main__":
    model, visual_model = create_floating_base("base")

    left_arm = load_robot_description("ur3_description", root_joint=None)
    prefix_frames(left_arm.model, left_arm.visual_model, "left")
    left_arm_placement = pin.SE3.Identity()
    left_arm_placement.translation = np.array([0.3, 0.0, 0.0])
    model, visual_model = pin.appendModel(
        model,
        left_arm.model,
        visual_model,
        left_arm.visual_model,
        model.getFrameId("base"),
        left_arm_placement,
    )

    right_arm = load_robot_description("ur3_description", root_joint=None)
    prefix_frames(right_arm.model, right_arm.visual_model, "right")
    right_arm_placement = pin.SE3.Identity()
    right_arm_placement.translation = np.array([-0.3, 0.0, 0.0])
    model, visual_model = pin.appendModel(
        model,
        right_arm.model,
        visual_model,
        right_arm.visual_model,
        model.getFrameId("base"),
        right_arm_placement,
    )

    robot = pin.RobotWrapper(
        model,
        collision_model=visual_model,  # well...
        visual_model=visual_model,
    )
    viz = start_meshcat_visualizer(robot)

    base_task = FrameTask(
        "base",
        position_cost=1.0,  # [cost] / [m]
        orientation_cost=1.0,  # [cost] / [rad]
        lm_damping=0.1,
    )
    left_ee_task = FrameTask(
        "left_ee_link",
        position_cost=1.0,  # [cost] / [m]
        orientation_cost=1.0,  # [cost] / [rad]
        lm_damping=1.0,  # tuned for this setup
    )
    right_ee_task = FrameTask(
        "right_ee_link",
        position_cost=1.0,  # [cost] / [m]
        orientation_cost=1.0,  # [cost] / [rad]
        lm_damping=1.0,  # tuned for this setup
    )
    posture_task = PostureTask(
        cost=1e-3,  # [cost] / [rad]
    )

    tasks = [base_task, left_ee_task, right_ee_task, posture_task]

    q_ref = custom_configuration_vector(
        robot,
        left_shoulder_lift_joint=1.0,
        left_shoulder_pan_joint=1.0,
        left_elbow_joint=1.0,
        right_shoulder_lift_joint=1.0,
        right_shoulder_pan_joint=1.0,
        right_elbow_joint=1.0,
    )
    configuration = pink.Configuration(robot.model, robot.data, q_ref)
    for task in tasks:
        task.set_target_from_configuration(configuration)
    viz.display(configuration.q)

    viewer = viz.viewer
    meshcat_shapes.frame(viewer["base"], opacity=1.0)
    meshcat_shapes.frame(viewer["base_target"], opacity=0.5)
    meshcat_shapes.frame(viewer["left_ee"], opacity=1.0)
    meshcat_shapes.frame(viewer["left_ee_target"], opacity=0.5)
    meshcat_shapes.frame(viewer["right_ee"], opacity=1.0)
    meshcat_shapes.frame(viewer["right_ee_target"], opacity=0.5)

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "quadprog" in qpsolvers.available_solvers:
        solver = "quadprog"

    rate = RateLimiter(frequency=200.0)
    dt = rate.period
    t = 0.0  # [s]
    while True:
        # Update task targets
        base_target = base_task.transform_target_to_world
        base_target.translation[2] = 0.3 * np.sin(2.0 * t)
        left_ee_target = left_ee_task.transform_target_to_world
        left_ee_target.translation[1] = 0.5 + 0.1 * np.sin(2.0 * t)
        left_ee_target.translation[2] = 0.2
        right_ee_target = right_ee_task.transform_target_to_world
        right_ee_target.translation[1] = 0.5 + 0.1 * np.sin(2.0 * t)
        right_ee_target.translation[2] = 0.2

        # Update visualization frames
        viewer["base_target"].set_transform(base_target.np)
        viewer["base"].set_transform(
            configuration.get_transform_frame_to_world(base_task.frame).np
        )
        viewer["left_ee_target"].set_transform(left_ee_target.np)
        viewer["left_ee"].set_transform(
            configuration.get_transform_frame_to_world(left_ee_task.frame).np
        )
        viewer["right_ee_target"].set_transform(right_ee_target.np)
        viewer["right_ee"].set_transform(
            configuration.get_transform_frame_to_world(right_ee_task.frame).np
        )

        # Compute velocity and integrate it into next configuration
        velocity = solve_ik(configuration, tasks, dt, solver=solver)
        configuration.integrate_inplace(velocity, dt)

        # Visualize result at fixed FPS
        viz.display(configuration.q)
        rate.sleep()
        t += dt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Ivan Domrachev, Simeon Nedelchev

"""Go2 squat with z-axis barrier."""

import meshcat_shapes
import numpy as np
import qpsolvers
from loop_rate_limiters import RateLimiter
import pinocchio as pin

import pink
from pink import solve_ik
from pink.barriers import PositionBarrier
from pink.tasks import FrameTask, PostureTask
from pink.visualization import start_meshcat_visualizer

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try ``pip install robot_descriptions``"
    ) from exc  # noqa: E501


if __name__ == "__main__":
    robot = load_robot_description(
        "go2_description", root_joint=pin.JointModelFreeFlyer()
    )
    viz = start_meshcat_visualizer(robot)

    q_ref = np.array(
        [
            -0.0,
            0.0,
            0.3,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.8,
            -1.57,
            0.0,
            0.8,
            -1.57,
            0.0,
            0.8,
            -1.57,
            0.0,
            0.8,
            -1.57,
        ]
    )

    configuration = pink.Configuration(robot.model, robot.data, q_ref)

    base_task = FrameTask(
        "base",
        position_cost=50.0,  # [cost] / [m]
        orientation_cost=1.0,  # [cost] / [rad]
    )

    posture_task = PostureTask(
        cost=1e-5,  # [cost] / [rad]
    )

    pos_barrier = PositionBarrier(
        "base",
        indices=[1, 2],
        p_max=np.array([0, 0.35]),
        gain=np.array([100.0, 100.0]),
        safe_displacement_gain=1.0,
    )
    barriers = [pos_barrier]

    tasks = [base_task, posture_task]

    for foot in ["FL_foot", "FR_foot", "RL_foot", "RR_foot"]:
        task = FrameTask(
            foot,
            position_cost=200.0,  # [cost] / [m]
            orientation_cost=0.0,  # [cost] / [rad]
        )
        tasks.append(task)

    for task in tasks:
        task.set_target_from_configuration(configuration)

    viewer = viz.viewer
    opacity = 0.5  # Set the desired opacity level (0 transparent, 1 opaque)

    meshcat_shapes.frame(viewer["base_target"], opacity=1.0)
    meshcat_shapes.frame(viewer["base"], opacity=1.0)

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "osqp" in qpsolvers.available_solvers:
        solver = "osqp"

    rate = RateLimiter(frequency=200.0)
    dt = rate.period
    t = 0.0  # [s]
    period = 4
    omega = 2 * np.pi / period
    while True:
        # Update task targets
        end_effector_target = base_task.transform_target_to_world
        phase = (t // period) % 2
        Ay = 0.1 * (1 - phase)
        Az = 0.2 * phase

        end_effector_target.translation[1] = 0.0 + Ay * np.sin(omega * t)
        end_effector_target.translation[2] = 0.3 + Az * np.sin(omega * t)

        # Update visualization frames
        viewer["base_target"].set_transform(end_effector_target.np)
        viewer["base"].set_transform(
            configuration.get_transform_frame_to_world(base_task.frame).np
        )

        velocity = solve_ik(
            configuration,
            tasks,
            dt,
            solver=solver,
            barriers=barriers,
        )
        configuration.integrate_inplace(velocity, dt)

        viz.display(configuration.q)
        rate.sleep()
        t += dt
